#!/usr/bin/env python

"""The Stud GA

References:
Khatib, Wael and Peter John Fleming. “The Stud GA: A Mini Revolution?” Parallel Problem Solving from Nature (1998).
"""

from random import random
from .population import HOFPopulation
from .utils import choice


class StudPopulation(HOFPopulation):

    params = {'fame_size': 2}
    
    def mate(self, mate_prob=None):
        """Mating in studGA

        individuals only mate with individuals in the hall of fame
        """

        mate_prob = mate_prob or self.mate_prob
        offspring = []
        for individual in self:
            if individual in self.halloffame:
                continue
            if random() < (mate_prob or self.mate_prob):
                other = choice(self.halloffame)
                offspring.append(individual.cross(other))
        self.extend(offspring)
        return offspring



#!/usr/bin/env python3

"""
Decorators
"""

from types import MethodType
from operator import methodcaller
import copy


def clear_cache(func):
    def mthd(obj, *args, **kwargs):
        result = func(obj, *args, **kwargs)
        obj.clear_cache()
        return result
    return mthd

def side_effect(func):
    """Decorator for methods with side effect

    Apply the decorator to methods with side effects.
    If all the methods called by a particular method have the decorator applied,
    it is not advisable to include the decorator in that method.
    
    Args:
        func (TYPE): a method
    
    Returns:
        Decorated method
    """

    def mthd(obj, *args, **kwargs):
        result = func(obj, *args, **kwargs)
        # clear the cache after calling the method
        if hasattr(obj, '_cache'):
            obj.clear_cache()
        return result
    return mthd


def clear_fitness(func):
    def mthd(obj, *args, **kwargs):
        result = func(obj, *args, **kwargs)
        obj.clear_cache('fitness')
        return result
    return mthd


usual_side_effect = ['mutate', 'extend', 'pop', 'remove', '__setitem__', '__setattr__', '__setstate__']


def method_cache(func, a):
    """cache for methods

    Pre-define `_cache` as an attribute of the obj.
    
    Args:
        func (TYPE): the original method
        a (TYPE): an attribute (or any value) computed by the method
    
    Returns:
        MethodType
    """
    def mthd(obj):
        # get the attribute from cache, otherwise compute it again
        if obj._cache[a] is None:
            f = obj.func()
            obj._cache[a] = f
            return f
        else:
            return obj._cache[a]
    return mthd


class add_cache:

    """Handle with cache for class
    
    Attributes:
        attrs (tuple[str]): a tuple of attributes what will be cached
        methods (tuple[str]): a tuple of method names that will be cached
        cmd (dict[str]): a dict of commands to handle with the cache
    """

    def __init__(self, attrs, methods=(), scope=None, cmd={}):
        self.methods = methods
        self.attrs = attrs
        self.scope = scope
        self.cmd = cmd

    def __call__(self, cls):

        # add `_cache` to the class
        if hasattr(cls, '_cache'):
            cls._cache.update({a: None for a in self.attrs})
        else:
            cls._cache = {a: None for a in self.attrs}

        @property
        def cache(obj):
            return obj._cache

        cls.cache = property(cache)

        def _clear_cache(obj, k=None):
            if k is None:
                obj._cache = {k: None for k in obj._cache.keys()}
            elif k in obj._cache:
                obj._cache[k] = None

        def _cleared(obj, k=None):
            if k is None:
                return all(v == None for v in obj._cache.values())
            elif k in obj._cache:
                return obj._cache[k] == None

        def _set_cache(obj, **d):
            obj._cache.update(d)

        cls.cleared = _cleared
        cls.clear_cache = _clear_cache
        cls.set_cache = _set_cache

        for n, c in self.cmd.items():
            def _c(obj):
                c(obj._cache)
            setattr(cls, n, _c)

        if hasattr(cls, 'copy'):
            cls_copy = cls.copy
            def _copy(obj, cache=True, *args, **kwargs):
                # set cache=True to copy the cache
                cpy = cls_copy(obj, *args, **kwargs)
                if cache:
                    cpy.set_cache(**obj._cache)
                return cpy
            cls.copy = _copy

        for a in self.attrs:
            if not hasattr(cls, a):
                raise AttributeError(f'The attribute "{a}" should be used in the algorithm!')
            def f(obj):
                """get the attribute from cache, 
                otherwise re-compute it by the default/parent method
                """
                if obj._cache[a] is None:      
                    v = getattr(super(cls, obj), a)
                    obj._cache[a] = v
                    return v
                else:
                    return obj._cache[a]
            setattr(cls, a, property(f))

        def _after_setter(obj):
            obj.clear_cache()

        cls.after_setter = _after_setter

        for name in self.methods:
            if hasattr(cls, name):
                setattr(cls, name, clear_cache(getattr(cls, name)))
            else:
                print(f'the class "{cls.__name__}" does not have method "{name}"')

        cls_new = cls.__new__
        def _new(cls, *args, **kwargs):
            obj = cls_new(cls, *args, **kwargs)
            obj._cache = copy.copy(cls._cache)
            return obj

        cls.__new__ = _new

        return cls


fitness_cache = add_cache(('fitness',))


class set_fitness:

    def __init__(self, f=None):
        self.f = f

    def __call__(self, cls):
        if self.f is None:
            if '_fitness' in globals():
                self.f = globals()['_fitness']
            else:
                raise Exception("""Function `_fitness` is not defined before setting fitness.
You may forget to create the class in the context of environment.""")
        cls._fitness = self.f
        return cls


class add_memory:

    """
    add the `_memory` dict to the cls/obj

    The memory dict stores the best solution,
    unlike the `_cache` dict which only records the last computing result.
    And it is not affected by the genetic operations.

    _memory {dict[str]} -- the memory for an object;
                        In general, the key is the property of the object.
    """

    def __init__(self, memory={}):
        self._memory = memory

    def __call__(self, cls):

        cls._memory = self._memory

        def memory(obj):
            return obj._memory

        cls.memory = property(memory)

        def _set_memory(obj, **d):
            obj._memory.update(d)

        cls.set_memory = _set_memory

        def fitness(obj):
            # get fitness from memory by default
            if obj.memory['fitness'] is None:
                return obj._fitness()
            else:
                return obj.memory['fitness']

        cls.fitness = property(fitness)

        for a in cls._memory:
            if a == 'fitness': continue
            def f(obj):
                """get the attribute from memory, 
                where the best solution is stored
                """
                if obj._memory[a] is None:      
                    v = getattr(super(cls, obj), a)
                    return v
                else:
                    return obj._memory[a]
            setattr(cls, a, property(f))

        if hasattr(cls, 'copy'):
            cls_copy = cls.copy
            def _copy(obj, *args, **kwargs):
                cpy = cls_copy(obj, *args, **kwargs)
                cpy._memory = copy.copy(obj._memory)
                return cpy
            cls.copy = _copy

        cls_new = cls.__new__
        def _new(cls, *args, **kwargs):
            obj = cls_new(cls, *args, **kwargs)
            obj._memory = copy.deepcopy(cls._memory)
            return obj

        cls.__new__ = _new

        return cls


def basic_memory(cls):
    # special case of `add_memory`

    cls = add_memory({'fitness':None, 'solution': None})(cls)

    def _backup(obj, check=True):
        """Backup the fitness and other information
        
        Args:
            check (bool, optional): check whether the fitness increases.
        """

        f = super(cls, obj).fitness
        if not check or (obj.memory['fitness'] is None or f > obj.memory['fitness']):
            obj.set_memory(fitness=f, solution=obj.solution)
            
    cls.backup = _backup

    def _init(obj):
        obj.backup(check=False)

    cls.init = _init

    return cls


usual_side_effect = ['mutate', 'extend', 'pop', 'remove', '__setitem__', '__setattr__', '__setstate__']

def method_cache(func, a):
    """cache for methods

    Pre-define `_cache` as an attribute of the obj.
    
    Args:
        func (TYPE): the original method
        a (TYPE): an attribute (or any value) computed by the method
    
    Returns:
        MethodType
    """

    def mthd(obj):
        # get the attribute from cache, otherwise compute it again
        if obj._cache[a] is None:
            f = obj.func()
            obj._cache[a] = f
            return f
        else:
            return obj._cache[a]
    return mthd


class regester_map:

    def __init__(self, mappings, map_=map):
        """
        Args:
            mappings (str, tuple of str): a mapping or mappings on the object
            map_ (None, optional): map method
        """

        if mappings:
            self.mappings = mappings
        else:
            raise Exception('Have not provided any mapping')
        self.map = map_

    def __call__(self, cls, map_=None):

        if map_ is None:
            if self.map is None:
                if hasattr(cls, map):
                    _map = cls.map
                else:
                    _map = map
            else:
                _map = self.map
        else:
            _map = map_

        # if type_:
        #     _map = lambda *args, **kwargs: type_(_map(**args, **kwargs))

        if isinstance(self.mappings, str):
            m = self.mappings
            def _m(obj, *args, **kwargs):
                return _map(methodcaller(m, *args, **kwargs), obj)
            setattr(cls, m, _m)
        elif isinstance(self.mappings, tuple):
            for m in self.mappings:
                def _m(obj, *args, **kwargs):
                    return _map(methodcaller(m, *args, **kwargs), obj)
                setattr(cls, m, _m)
        else:
            raise TypeError('`mappings` has to be an instance of str or a tuple of strings')

        return cls


class Regester:
    # regerster operators, used in the future version

    def __init__(name, key=None):
        self.name = name
        self.key = key

    def __call__(self, cls):

        def _regester_operator(obj, name, key=None):
            if hasattr(obj, name):
                raise AttributeError(f'"{name}" is an attribute of "{obj.__name__}", and would not be regestered.')
            if not hasattr(obj, '_operators'):
                obj._operators = [key]
            else:
                obj._operators.append(key)
            setattr(obj, name, MethodType(key, obj))

        def _element_regester(obj, name, e):
            if hasattr(obj, e):
                raise AttributeError(f'"{e}" is an attribute of "{obj.__name__}", would not be regestered.')
            obj._elements.append(e)
            setattr(obj, name, e)

        cls.regester_operator = _regester_operator
        cls.regester_element = _regester_element

        return cls



#!/usr/bin/env python3


import numpy as np
import numpy.linalg as LA



class KMeans(BaseProblem):
    """KMeans clustering Problem

    ERM:
    min J(c,mu) = sum_c sum_{x:c} ||x-mu_c||
    """
    def __init__(self, X, n_components=2):
        self.X = X
        self.n_components = n_components

    @staticmethod
    def random(N, p=2):
        X1 = np.random.normal(1, 1, (N, p))
        X2 = np.random.normal(2, 1, (N, p))
        X = np.vstack((X1, X2))
        return KMeans(X, n_components=2)

    def __call__(self, x):
        # xi = k  iff  Xi in k-class
        cs = set(x)
        xs = {c:[self.X[i] for i, k in enumerate(x) if k==c] for c in cs}
        J = np.mean([np.sum([LA.norm(xi - np.mean(x, axis=0)) for xi in x]) for c, x in xs.items()])
        return J

# from scipy.stats import norm

# class MixGaussian(BaseProblem):
#     """Mix Gaussian clustering Problem
#     X ~ sum a_i p(x|mu_i, S_i)

#     max L(ci,{mu_i,S_i})= prod_k p(xk|ci, {mu_i, S_i}) = prod_k sum_i a_i p(xk|mu_i, S_i)
#     """
#     def __init__(self, X, n_components=2):
#         self.X = X
#         self.n_components = n_components

#     @staticmethod
#     def random(N, p=2):
#         X1 = norm.rvs(size=(N, p))
#         X2 = norm.rvs(loc=2, size=(N, p))
#         X = np.vstack((X1, X2))
#         return MixGaussian(X, n_components=2)

#     def __call__(self, t):
#         cs = set(t)
#         xs = {c:[self.X[i] for i, k in enumerate(x) if k==c] for c in cs}
#         # mus = {c:np.mean(x, axis=0) for c, x in xs.items()}
#         J = np.sum([np.prod([norm.pdf(xi, ti) for xi in x]) for c, x in xs.items()])
#         return J



#!/usr/bin/env python3

"""
Self Adaptive GA
"""


from random import random
from . import MixedIndividual, StandardPopulation, BinaryChromosome, FloatChromosome


class BaseSelfAdaptiveIndividual(MixedIndividual):

    @property
    def mutate_prob(self):
        raise NotImplementedError

    @mutate_prob.setter
    def mutate_prob(self, v):
        raise NotImplementedError

    @property
    def mate_prob(self):
        raise NotImplementedError

    @mate_prob.setter
    def mate_prob(self, v):
        raise NotImplementedError

    @property
    def desire(self):
        raise NotImplementedError


class SelfAdaptiveIndividual(BaseSelfAdaptiveIndividual):
    """Individual for Self-adaptive GA

    Provide at least 2 chromosomes, one of them is coded by float numbers
    representing the prob of mutation and mating.

    Extends:
        BaseSelfAdaptiveIndividual
    """
    
    # By default, the second chromosome (a float vector) encodes the prob of mutation and mating

    element_class = BinaryChromosome, FloatChromosome

    def mutate(self):
        if random() < self.mutate_prob:
            super().mutate()

    def mate(self, other, mate_prob=None):
        if random() < (mate_prob or (self.mate_prob + other.mate_prob)/2):
            return super().mate(other)
        else:
            return self

    @property
    def trait(self):
        return self.chromosomes[-1]

    @property
    def mutate_prob(self):
        return self.trait[0]

    @mutate_prob.setter
    def mutate_prob(self, v):
        self.trait[0] = v

    @property
    def mate_prob(self):
        return self.trait[1]

    @mate_prob.setter
    def mate_prob(self, v):
        self.trait[1] = v

    @property
    def desire(self):
        return self.trait[2]
    

class RankingIndividual(SelfAdaptiveIndividual):
    ranking = None

    # def mate(self, other, mate_prob=None):
    #     if other.ranking:
    #         if self.threshold <= other.ranking:
    #             return super().mate(other, mate_prob=1)
    #         elif self.threshold <= 2* other.ranking:
    #             return super().mate(other, mate_prob=0.8)
    #         else:
    #             return super().mate(other, mate_prob=0.5)
    #     else:
    #         return super().mate(other)

    # @property
    # def threshold(self):
    #     return self.trait[-1]


def lim(r, e):
    e -= 0.0001
    return 0 if r <= e else (r - e)**2 / (1 - e)


class SSAPopulation(StandardPopulation):
    # super self-adaptive population

    def transition(self, *args, **kwargs):

        self.mate()
        self.mutate()
        self.doom()
        
        for individual in self:
            individual.age += 1
        if self.is_crowd():
            self.select(0.3)

    def doom(self):
        # Remove the dead individuals
        self.individuals = [individual for individual in self if not individual.is_dead()]

    def mate(self):
        self.rank()
        children = []
        for individual, other in zip(self[:-1], self[1:]):
            if random() < min(individual.cross_prob, other.cross_prob) and self.match(individual, other):
                children.append(individual.cross(other))
        self.extend(children)
        return children

    @classmethod
    def match(cls, individual, other):
        if individual.label != other.label:
            a, b, c = lim(other.ranking, individual.expect), lim(individual.ranking, other.expect), abs(individual.age - other.age)
            p = 1 - c/(20*min(a, b)+1)
            return random() < p
        else:
            return random() < 0.05

    def is_crowd(self):
        return len(self) > 8 * self.default_size

    def select(self, p=0.5):
        # self.select(n_sel=p)
        self.individuals = choice_with_fitness(self.individuals, fs=None, n=int(self.n_individuals*p))
        # self.individuals = [individual for individual in self.individuals if random() < p]



#!/usr/bin/env python3


"""Evolution Programming

Invented by L. Fogel[1966] for designing FSM initially.

General form of the algorithm:

    initialize a population with N individuals
    loop:
        calculate f(x) for each x in population
        mutate x for each x
        get new population (mixed with the original population)
        select best N individuals from the 2N mixed population

    The mutation:
        x' <- x + r*sqrt(v)
        v' <- v + c*r*sqrt(v) (make sure that v>epsilon)

Remark:
    No cross operation in EP
"""

from random import choice
from toolz.itertoolz import groupby
from operator import attrgetter

import numpy as np

from .base import PopulationMixin, BaseChromosome
from .chromosome import FloatChromosome
from .individual import MixedIndividual

from .deco import side_effect


class BaseEPIndividual(MixedIndividual):
    """Base class of EP Individual Class

    A single solution in EP
    """

    params = {'c':0.1, 'epsilon':0.0001}

    element_class = BaseChromosome, FloatChromosome

    def decode(self):
        return self.chromosomes[0]

    @property
    def variance(self):
        return self.chromosomes[1]

    @variance.setter
    def variance(self, v):
        self.chromosomes[1] = v
    
    @side_effect
    def mutate(self):
        rx = np.random.randn(*self.chromosomes[0].shape)
        self.chromosomes[0] += rx * np.sqrt(self.variance)

        rv = np.random.randn(*self.variance.shape)
        self.variance += self.c * rv * np.sqrt(self.variance)
        self.variance = np.maximum(self.variance, self.epsilon)


class EvolutionProgramming(PopulationMixin):
    """Evolution Programming
    
    Extends:
        PopulationMixin
    """
    
    element_class = BaseEPIndividual

    def select(self):
        self.sort()
        d = groupby(attrgetter('fitness'), self)
        inds = []
        ks = np.sort(list(d.keys()))[::-1]
        flag = True
        while flag:
            for k in ks:
                if d[k]:
                    a = choice(d[k])
                    inds.append(a)
                    d[k].remove(a)
                    if len(inds) <= self.default_size:
                        flag = False
                        break
        self.individuals = inds

    def transition(self, *args, **kwargs):
        cpy = self.clone()
        self.mutate()
        self.merge(cpy)
        self.select()
    



#!/usr/bin/env python3

"""
Spcical functions for benchmark
"""

from math import sin, sqrt, pi
import numpy as np

def rosenbrock(x:np.ndarray):
    x = np.asarray(x)
    return np.sum( (100 * (x[1:]-x[:-1]**2)**2 + x[:-1]-1) ** 2 )


def schaffer(x:np.ndarray):
    r2 = np.sum(x**2)
    return (sin(sqrt(r2)) ** 2 - 0.5) / (1+ 0.001*r2**2)**2


def rastrigrin(x:np.ndarray):
    return np.sum(x**2 - 10 * np.cos(2*pi*x)) + 40


def griewangk(n=5):
    def f(x:np.ndarray):
        return np.sum(x**2)/4000 - np.prod(np.cos(x / np.sqrt(np.arange(1, n+1))))
    return f


def michalewiez(n=5):
    def f(x:np.ndarray):
        return - sum(sin(x) * sin(np.arange(1, n+1) *x**2/pi)**20)/4000
    return f


def hansen(n=5):
    def f(x:np.ndarray):
        return np.sum(np.arange(1, n+1)*(np.cos(np.arange(n))*x+1)) * np.sum(np.arange(1, n+1)*(np.cos(np.arange(2, n+2))*x+1))
    return f


def alpine(x):
    return np.sum(np.abs((np.sin(x) + 0.1) * x))



#!/usr/bin/env python3


from itertools import product
import threading

from .base import BasePopulation, BaseMultiPopulation
from .utils import *


class MultiPopulation(BaseMultiPopulation):

    def mate(self):
        for p in self:
            p.mate()


class DualPopulation(MultiPopulation):

    """Multi-population composed by male and female population
    """
    
    params = {'n_elders':0.5, 'mate_prob':0.75}

    default_size = 2

    @property
    def male_population(self):
        return self.populations[0]

    @property  
    def female_population(self):
        return self.populations[1]

    @property  
    def males(self):
        return self.populations[0].individuals

    @property  
    def females(self):
        return self.populations[1].individuals

    @property
    def male_fitness(self):
        return self.populations[0].fitness

    @property
    def female_fitness(self):
        return self.populations[1].fitness

    def mate(self):
        self.populations[0].rank(tied=True)
        self.populations[1].rank(tied=True)
        children = []
        def _target(male, female):
            if random()<0.5:
                child = male.cross(female)
            else:
                child = female.cross(male)
            children.append(child)
        ps = [threading.Thread(target=_target, args=(male, female)) for male, female in product(self.males, self.females)
            if random() < self.mate_prob and self.match(male, female)]

        for p in ps:
            p.start()
        for p in ps:
            p.join()

        self.populations[0].extend(children[::2])
        self.populations[1].extend(children[1::2])

    def match(self, male, female):
        return True

    def transition(self, *args, **kwargs):
        elder = self.__class__([
            self.populations[0].get_best_individuals(self.n_elders * self.populations[0].default_size, copy=True),
            self.populations[1].get_best_individuals(self.n_elders * self.populations[1].default_size, copy=True)
            ])
        self.select()
        self.mate()
        self.mutate()
        self.merge(elder)

    def select(self):
        self.populations[0].select()
        self.populations[1].select()

    def mutate(self):
        self.populations[0].mutate()
        self.populations[1].mutate()

    def merge(self, other):
        self.populations[0].merge(other.populations[0])
        self.populations[1].merge(other.populations[1])


class HybridPopulation(MultiPopulation):

    def migrate(self, migrate_prob=None, copy=True):
        migrate_prob = migrate_prob or self.migrate_prob
        for this, other in zip(self[:-1], self[1:]):
            if random() < migrate_prob:
                if isinstance(this, self.element_class):
                    this_best = this.get_best_individual(copy=copy)
                    if isinstance(other, self.element_class):
                        other.append(this.get_best_individual(copy=copy))
                        this.append(other.get_best_individual(copy=copy))
                    else:
                        this.append(other.copy())
                else:
                    this_best = this.copy()
                    if isinstance(other, self.element_class):
                        other.append(this.get_best_individual(copy=copy))




#!/usr/bin/env python3


"""
Mixin classes for iterative algorithms or models,
where the basic operations are defined.

IterativeMixin: base class for all iterative algorithms
FitnessMixin: IterativeMixin with `fitness`
CollectiveMixin: base class for all iterative algorithms with multi-objects
PopulationMixin: subclass of FitnessMixin, population-like iterative algorithms

Relation of the classes:
IterativeMixin  --->  CollectiveMixin
    |                      |
    |                      |
    v                      v
FitnessMixin  --->  PopulationMixin
"""


from operator import methodcaller, attrgetter

import numpy as np
import pandas as pd

try:
    from ezstat import Statistics
except:
    from ._stat import Statistics

from .deco import side_effect

from .errors import *


class IterativeMixin:
    # Mixin class for iterative algrithms

    params = {'n_iter': 100}

    # @property
    # def _row(self):
    #     best = self.solution
    #     return f'{best} & {best.fitness}'

    def init(self):
        pass

    def transition(self, *args, **kwargs):
        """The core method of the object.

        This method transitions one state of the object to another state
        based on certain rules, such as crossing and mutating for individuals in GA,
        or the moving method in Simulated Annealing.
        """
        raise NotImplementedError('`transition` (the core of the algorithm) is not defined yet!')

    def local_search(self, *args, **kwargs):
        """
        The local search method for a global search algorithm.
        """
        raise NotImplementedError('If you apply a local search algorithm, you must define the `local_search` method.')

    def ezolve(self, n_iter=None, init=True):
        # Extreamly eazy evolution method for lazybones
        n_iter = n_iter or self.n_iter
        if init:
            self.init()
        for k in range(1, n_iter+1):
            self.transition(k)

    def evolve(self, initialize:bool=True, n_iter:int=100, period:int=1, verbose:bool=False, history=False, stat=None, attrs=('solution',), control=None):
        """Get the history of the whole evolution

        Keyword Arguments:
            n_iter {number} -- number of iterations (default: {None})
            period {integer} -- the peroid of stat
            verbose {bool} -- to print the iteration process
            stat {dict} -- a dict(key: function mapping from the object to a number) of statistics 
                           The value could be a string that should be a method pre-defined.
            history {bool|DataFrame} -- True for recording history, or a DataFrame object recording previous history.
            attrs {tuple[str]} -- attributes of the object
        
        Returns:
            DataFrame | None
        """

        assert control is None or callable(control)
 
        n_iter = n_iter or self.n_iter

        if isinstance(stat, dict): stat = Statistics(stat)
        
        if initialize:
            self.init()

        if history is True:
            res = stat(self)
            history = pd.DataFrame(data={k:[v] for k, v in res.items()})
            history_flag = True
        elif history is False:
            history_flag = False
        elif isinstance(history, pd.DataFrame):
            history_flag = True
        else:
            raise TypeError('The argument `history` should be an instance of `pandas.DataFrame` or `bool`.')
        # n_iter = n_iter or self.n_iter
        if verbose:
            from toolz.itertoolz import concat
            if not history_flag:
                res = stat(self)
            print(f"""
{" & ".join(concat((("iteration",), attrs, res.keys())))}
-------------------------------------------------------------
{" & ".join(map(str, concat((("[0]",), (getattr(self, attr) for attr in attrs), res.values()))))}""")

        for t in range(1, n_iter+1):
            self.transition(t)
            if history_flag and (period == 1 or t % period ==0):
                res = stat(self)
                history = pd.concat([history,
                    pd.Series(res.values(), index=res.keys()).to_frame().T],
                    ignore_index=True)
            if verbose and (period == 1 or t % period ==0):
                print(f'{" & ".join(map(str, concat((("[%d]"%t,), (getattr(self, attr) for attr in attrs), res.values()))))}')

            if control:
                if control(self):
                    # if it satisfies the control condition
                    break
        return history

    def perf(self, n_repeats=10, timing=True, *args, **kwargs):
        """Get performance of Algo.

        Keyword Arguments:
            n_repeats {number} -- number of repeats to running algo. (default: {10})
            timing {bool} -- measure execution the time
        
        Returns:
            history, time
        """

        from time import process_time

        times = []
        data = None
        for _ in range(n_repeats):
            cpy = self.copy(cache=False)
            data0 = cpy.evolve(verbose=False, *args, **kwargs)
            if timing:
                start = process_time()
                data0 = cpy.evolve(verbose=False, *args, **kwargs)
                end = process_time()
                delta = end - start
            times.append(delta)
            if data is None:
                data = data0
            else:
                data += data0

        return data / n_repeats, np.mean(times)

    def copy(self, *args, **kwargs):
        raise NotImplementedError

    def clone(self):
        return self.copy()

    def decode(self):
        return self

    @classmethod
    def encode(cls, x):
        # encode x to a chromosome
        raise NotImplementedError

    @property
    def solution(self):
        return self.decode()
 
    def save(self, filename=None, check=False):
        """Save the object to file using pickle
        
        Args:
            filename (None, optional): the path of the pickle file
            check (bool, optional): check whether the file has existed.
        
        Raises:
            FileExistsError: Description
        """
        import pickle, pathlib
        if filename is None:
            filename = f'{self.__class__.__name__}-{self.__name__}'
        if isinstance(filename, str):
            pklPath = pathlib.Path(filename).with_suffix('.pkl')
        if check and pklPath.exists():
            raise FileExistsError(f'File {filename} has existed!')
        with open(pklPath, 'wb') as fo:
            pickle.dump(self, fo)

    @classmethod
    def load(cls, filename=None):
        """Load the object from the pickle file
        
        Args:
            filename (None, optional): the path of the pickle file
        
        Returns:
            TYPE: the object of `cls`

        Raises:
            FileNotFoundError: Do not find the file
        """
        import pickle, pathlib
        if filename is None:
            filename = f'{cls.__name__}-{pop}'
        if isinstance(filename, str):
            pklPath = pathlib.Path(filename).with_suffix('.pkl')
        if pklPath.exists():
            with open(pklPath, 'rb') as fo:
                obj = pickle.load(fo)
                if not isinstance(obj, cls):
                    raise TypeError('The object loaded is not an instance of "{cls}".')
        else:
            raise FileNotFoundError(f'Could not find the file {filename}!')

    def after_setter(self):
        if hasattr(self, '_cache'):
            self.clear_cache()

    @classmethod
    def solve(cls, *args, **kwargs):
        return cls.random().evolve(*args, **kwargs).solution


class FitnessMixin(IterativeMixin):
    """Iterative models drived by the fitness/objective function

    The fitness should be stored until the the state of the model is changed.
    
    Extends:
        IterativeMixin
    """

    def get_fitness(self):
        raise NotImplementedError

    def _fitness(self):
        raise NotImplementedError

    @property
    def fitness(self):
        return self._fitness()

    @classmethod
    def set_fitness(cls, f=None, decode=None):
        """Set fitness computing method
        
        Args:
            f (None, optional): function to evalute the fintess
            decode (None, optional): decode the individual or not before calcuating fitness
                                    (deprecated currently, eps. when you use `memory`!)
        
        Returns:
            An individual class with the fitness `f`
        """

        if f is None:
            if '_fitness' in globals():
                f = globals()['_fitness']
            else:
                raise Exception('Function `_fitness` is not defined before setting fitness. You may forget to create the class in the context of environment.')
        if not decode:
            def _fitness(obj):
                return f(obj)
        else:
            def _fitness(obj):
                return f(obj.decode())
        cls._fitness = _fitness
        return cls

    def evolve(self, stat=None, *args, **kwargs):
        """Get the history of solution and its fitness by default.
        """

        if stat is None:
            stat = {'Fitness': 'fitness'}
        return super().evolve(stat=stat, *args, **kwargs)

    def copy(self, type_=None, element_class=None, *args, **kwargs):
        type_ = type_ or self.__class__
        element_class = element_class or type_.element_class
        if isinstance(type_.element_class, tuple):
            return type_([c.copy(type_=t) for c, t in zip(self, type_.element_class)])
        else:
            return type_([c.copy(type_=element_class) for c in self])

    def clone(self):
        return self.__class__(list(map(methodcaller('clone'), self)))

    # def diff_fitness(self):
    #     return self.fitness - self.previous_fitness

    # def previous_fitness(self):
    #     return self.memory['fitness']


class CollectiveMixin(IterativeMixin):
    # mixin class for swarm intelligent algorithm

    map = map

    def init(self, *args, **kwargs):
        for element in self:
            if hasattr(element, 'init'):
                element.init(*args, **kwargs)

    def transition(self, *args, **kwargs):
        for element in self:
            element.transition(*args, **kwargs)

    @side_effect
    def remove(self, individual):
        self.elements.remove(individual)

    @side_effect
    def pop(self, k=-1):
        self.elements.pop(k)

    @side_effect
    def extend(self, inds):
        self.elements.extend(inds)

    @side_effect
    def append(self, ind):
        self.elements.append(ind)

    def random_select(self, n_sel=None, copy=False):
        if n_sel is None:
            k = np.random.randint(len(self))
            return self[k]
        else:
            ks = np.random.randint(len(self), size=n_sel)
            return self[ks]


class PopulationMixin(FitnessMixin, CollectiveMixin):
    """mixin class for population-based heuristic algorithm

    It is consisted of a collection of solutions.
    """

    def evolve(self, stat=None, *args, **kwargs):
        """Get the history of the whole evolution
        """

        if stat is None:
            stat = {'Best Fitness': 'max_fitness', 'Mean Fitness': 'mean_fitness',
            'STD Fitness': 'std_fitness'}
        return super().evolve(stat=stat, *args, **kwargs)

    @classmethod
    def set_fitness(cls, *args, **kwargs):
        # set fitness for the element_class.
        if hasattr(cls.element_class, 'set_fitness'):
            cls.element_class.set_fitness(*args, **kwargs)
        else:
            raise AttributeError(f'{cls.element_class} does not have `set_fitness`')
        return cls

    @property
    def fitness(self):
        # The fitness of the entire population is the maximum fitness of the individuals
        return self.max_fitness

    # def _fitness(self):
    #     """Calculate the fitness of the whole population

    #     Fitness of a population is the best fitness by default.
    #     (not recommended to be overridden)
    #     """
    #     return self.max_fitness

    def get_all_fitness(self):
        return list(self.map(attrgetter('fitness'), self))

    def get_all(self, key='fitness'):
        return list(self.map(attrgetter(key)), self)

    @property
    def mean_fitness(self):
        return np.mean(self.get_all_fitness())

    @property
    def std_fitness(self):
        return np.std(self.get_all_fitness())

    @property
    def best_fitness(self):
        print('`best_fitness` is depricated and please use `max_fitness`')
        return np.max(self.get_all_fitness())

    @property
    def max_fitness(self):
        return np.max(self.get_all_fitness())

    @property
    def stat_fitness(self, s=np.max):
        f = self.get_all_fitness()
        if isinstance(s, tuple):
            return tuple(si(f) for si in  s)
        else:
            return s(f)

    def get_best_element(self, copy=False):
        """Get best element
        
        Args:
            copy (bool, optional): return the copy of the selected element, if `copy=True`
        
        Returns:
            An element
        """

        k = np.argmax(self.get_all_fitness())
        if copy:
            return self[k].copy()
        else:
            return self[k]

    def get_best_elements(self, n=1, copy=False):
        """Get first n best elements
        
        Args:
            n (int, optional): the number of elements selected
            copy (bool, optional): if copy=True, then it returns the copies of elements
        
        Returns:
            n elements
        """

        if n < 1:
            n = int(self.n_elements * n)
        elif not isinstance(n, int):
            n = int(n)

        if copy:
            return [self[k].copy() for k in self.argsort()[-n:]]
        else:
            return [self[k] for k in self.argsort()[-n:]]

    @property
    def best_element(self):
        """Get the best element

        The difference between `best_element` and `get_best_element` is that
        `best_element` only returns the reference of the selected element.
        
        Returns:
            The best element
        """

        k = np.argmax(self.get_all_fitness())
        return self[k]

    @property
    def solution(self):
        return self.best_element.decode()

    def get_worst_element(self, copy=False):
        # see get_best_element
        k = np.argmin(self.get_all_fitness())
        if copy:
            return self[k].copy()
        else:
            return self[k]

    def get_worst_elements(self, n=1, copy=False):
        # see get_best_elements
        if n < 1:
            n = int(self.n_elements * n)
        elif not isinstance(n, int):
            n = int(n)

        if copy:
            return [self[k].copy() for k in self.argsort()[:n]]
        else:
            return [self[k] for k in self.argsort()[:n]]

    @property
    def worst_element(self):
        k = np.argmin(self.get_all_fitness())
        return self[k]

    def sorted_(self):
        # return a list of sorted individuals
        return [self[k] for k in self.argsort()]

    def sort(self):
        # sort the whole population
        ks = self.argsort()
        self.__elements = self[ks]

    def argsort(self):
        return np.argsort(self.get_all_fitness())

    def drop(self, n=1):
        if n < 1:
            n = int(self.n_elements * n)
        elif not isinstance(n, int):
            n = int(n)
        ks = self.argsort()
        self.elements = [self[k] for k in ks[n:]]

    def observe(self):
        for o in self:
            o._population = self



#!/usr/bin/env python3

"""
Optimization Helpers for GA.

Users can use the following example to optimize a multivariate function by GA directly
without encoding the solutions to chromosomes or individuals.

Example:

    ```
    # min x1^2+x2, x1,x2 in (-1,1)
    ga_minimize(lambda x:x[0]**2+x[1], (-1,1), (-1,1))
    ```
"""


import numpy as np

from .chromosome import BinaryChromosome
from .individual import makeIndividual, makeBinaryIndividual
from .population import StandardPopulation
from .de import DifferentialEvolution


def _decode(c, a, b):
    """Decode a binary sequence to a real number in [a,b]
    
    Args:
        c (binary seqence): Description
        a (number): lower bound
        b (number): upper bound
    
    Returns:
        number: a number in [a, b], represented by binary sequence,
        that 00...0 corresponds a, and 11...1 corresponds b
    
    Raises:
        e: fail to import a 3rd-part lib
    """

    try:
        from digit_converter import IntervalConverter
    except ImportError as e:
        print('The default decoder requires the 3rd-part lib `digit_converter`')
        raise e
    return IntervalConverter(a, b)(c)


def _make_individual(func, *xlim, size=8, decode=_decode):
    """Make an individual class
    
    Args:
        *xlim: intervals
        size (int, tuple of int): the length of the encoding of xi
    """

    ndim = len(xlim)

    if ndim == 1:
        class _Individual(BinaryChromosome // size):

            def decode(self):
                return decode(self, *xlim[0])
    else:
        class _Individual(makeIndividual(n_chromosomes=ndim, size=size)):
            default_size = ndim

            def decode(self):
                return np.asarray([decode(c, a, b) for c, (a, b) in zip(self, xlim)])

    return _Individual.set_fitness(lambda obj: - func(obj.decode()))


def ga_minimize(func, *xlim, decode=_decode, population_size=20, size=8, **kwargs):
    """
    GA(with hall of fame) for minimizing the function `func` defined on `xlim`

    Arguments:

        func {function}: objective function defined on R^n
        xlim {tuple of number pairs}: the intervals of xi
        decode {mapping}: transform a binary sequence to a real number
            ('0-1' sequence, lower_bound, upper_bound) |-> xi
        population_size {int}: size of the population
        size {int or tuple of int}: the length of the encoding of xi

    Example:

        ga_minimize(lambda x:x[0]**2+x[1], (-1,1), (-1,1))
    """

    _Individual = _make_individual(func, *xlim, size=size, decode=decode)
    _Population = StandardPopulation[_Individual] // population_size

    pop = _Population.random()
    pop.ezolve(**kwargs)

    return pop.solution


def de_minimize(func, *xlim, decode=_decode, population_size=20, size=8, **kwargs):
    """
    DE for minimizing the function `func` defined on `xlim`

    Arguments:

        func {function}: objective function defined on R^n
        xlim: the intervals of xi
        decode: transform a binary sequence to a real number
            ('0-1' sequence, lower_bound, upper_bound) |-> xi
        population_size: size of the population
        size: the length of the encoding of xi

    Example:

        de_minimize(lambda x:x[0]**2+x[1], (-1,1), (-1,1))
    """

    _Individual = _make_individual(func, *xlim, size=size, decode=decode)
    _Population = DifferentialEvolution[_Individual] // population_size

    pop = _Population.random()
    pop.ezolve(**kwargs)
    return pop.solution


def ga_minimize_1d(func, xlim, decode=_decode, population_size=20, size=8, *args, **kwargs):
    """
    GA(with hall of fame) for minimizing 1D function `func` defined on the interval `xlim`

    Arguments:
        func: objective function defined on R
        xlim {pair of numbers}: the interval of x
        decode: transform a binary sequence to a real number
            ('0-1' sequence, lower_bound, upper_bound) |-> xi
        population_size {int}: size of the population
        size {int}: the length of the encoding of x

    Example:
        ga_minimize_1d(lambda x:x**2, (-1,1))
    """

    _Chromosome = _make_individual(func, xlim, size=size, decode=decode)
    _Population = StandardPopulation[_Chromosome] // population_size

    pop = _Population.random()
    pop.ezolve(**kwargs)
    return pop.solution


class Optimizer:

    """Optimizer class for optimization problem
    
    Attributes:
        min_max (str): 'min' or 'max'
        Population: GA population or other evolutionary algorithms
    """
    
    def __init__(self, Population=None, min_max='min'):
        self.Population = Population
        self.min_max = min_max

    def __call__(self, func, *xlim, decode=_decode, population_size=20, size=8, **kwargs):

        _Individual = _make_individual(func, *xlim, size=size, decode=decode)
        _Population = self.Population[_Individual] // population_size

        if self.min_max == 'min':
            def _evaluate(obj):
                return - func(obj.decode())
        else:
            def _evaluate(obj):
                return func(obj.decode())
        self.Population.set_fitness(_evaluate)

        pop = self.Population.random()
        pop.ezolve(**kwargs)
        return pop.solution




#!/usr/bin/env python3


from .base import *
from .mixin import *
from .errors import *
from .gene import *
from .chromosome import *
from .individual import *
from .population import *
from .multipopulation import *

# deprecated in future
from .saga import *
from .studga import *
from .pso import *
from .es import *
from .ep import *
from .de import *
from .ba import *


__version__ = "1.6"

__template__ = """
from pyrimidine.chromosome import BinaryChromosome
from pyrimidine.individual import MonoIndividual, binaryIndividual
from pyrimidine.population import StandardPopulation

from pyrimidine.benchmarks.optimization import *

n = 50
_evaluate = Knapsack.random(n)


# Define individual
class MyIndividual(MonoIndividual):
    element_class = BinaryChromosome//n

    def _fitness(self) -> float:
        # To evaluate an individual!
        return _evaluate(self.chromosome)

# Equiv. to
# MyIndividual = MyIndividual[BinaryChromosome//n].set_fitness(lambda o: _evaluate(o.chromosome))
# MyIndividual = (BinaryChromosome//n).set_fitness(_evaluate) # Algebraically equiv.

# Define Population
class MyPopulation(HOFPopulation):
    element_class = MyIndividual
    default_size = 20

# Equiv. to
# MyPopulation = StandardPopulation[MyIndividual] // 20

# Define MyPopulation as a container of BinaryChromosome
# MyPopulation = StandardPopulation[BinaryChromosome//n].set_fitness(_evaluate) // 20

pop = MyPopulation.random(size=n)

stat={'Mean Fitness': 'fitness', 'Best Fitness': 'max_fitness',
  'Standard Deviation of Fitnesses': 'std_fitness'}
data = pop.evolve(stat=stat, n_iter=200, history=True)


import matplotlib.pyplot as plt
fig = plt.figure()
ax = fig.add_subplot(111)
ax2 = ax.twinx()
data[['Mean Fitness', 'Best Fitness']].plot(ax=ax)
ax.legend(loc='upper left')
data['Standard Deviation of Fitnesses'].plot(ax=ax2, style='r--')
ax2.legend(loc='lower right')
ax.set_xlabel('Generations')
ax.set_ylabel('Fitness')
plt.show()
"""



#!/usr/bin/env python3

import numpy as np
import numpy.linalg as LA


def lsq(X, A, B, alpha=0.1):
    return LA.norm(np.dot(A,X)-B) + alpha * LA.norm(X)


fun = lambda x: (1 - x[0]) ** 2 + 100 * (x[1] - x[0] ** 2) ** 2
bnds = ((-2, 2), (-1, 3))




#!/usr/bin/env python3


"""
Metaclasses
"""

import copy
from types import MethodType
from collections.abc import Iterable
from operator import attrgetter, methodcaller


def inherit(attrs, attr, bases):
    """Inherit attribute `attr` from `bases`
    
    Args:
        attrs (dict): The attribution dictionary of an object 
        attr (string): An attribution whose value is a dict
        bases (tuple): The base classes
    
    Returns:
        dict: The updated dictionary of attributions
    """

    v = {}
    for b in bases[::-1]:
        if hasattr(b, attr):
            v.update(getattr(b, attr))

    v.update(attrs.get(attr,  {}))
    # attributes in `v` should not cover the attributes of the object
    v = {k:vk for k, vk in v.items() if k not in attrs}
    attrs[attr] = v
    return attrs


class ParamType(type):
    """Just a wrapper of `type`

    The key-value pairs in `params` could be inherited from super classes, like attributes.
    It make users set and manage parameters of classes or instances more conveniently.
    """

    def __new__(cls, name, bases=(), attrs={}):

        # inherit alias instead of overwriting it, when setting `alias` for a subclass
        # `alias` is not recommended to use.
        attrs = inherit(attrs, 'alias', bases)

        # inherit params instead of overwriting it, when setting `params` for a subclass
        attrs = inherit(attrs, 'params', bases)

        def _getattr(self, key):
            if key in self.params:
                return self.params[key]
            elif key in self.alias:
                return getattr(self, self.alias[key])
            else:
                raise AttributeError(f"""`{key}` is neither an attribute of the object of `{self.__class__}`, nor in `params` or `alias`;
                    If you are sure that `{key}` has been defined as an attribute, then you should check the definition statement.
                    If there is no syntax problem, then it is probably about the type error in the statement.""")
        attrs['__getattr__'] = _getattr

        def _setattr(self, key, value):
            if key in self.__dict__:
                self.__dict__[key] = value
            elif key in self.params:
                self.params[key] = value
            elif key in self.alias:
                setattr(self, self.alias[key], value)
            else:
                object.__setattr__(self, key, value)
        attrs['__setattr__'] = _setattr

        # if 'check' in attrs:
        #     attrs['check'](cls)

        return super().__new__(cls, name, bases, attrs)

    @classmethod
    def __prepare__(cls, name, bases):
        return {"alias":{}, "params":{}}

    def set(self, *args, **kwargs):
        # set the attributes dynamically
        for k in args:
            if k in globals():
                setattr(self, k, globals()[k])
            else:
                raise NameError(f"name '{k}' is not defined.")
        for k, v in kwargs.items():
            setattr(self, k, v)
        return self

    def set_params(self, **kwargs):
        # set `params` dynamically
        self.params.update(kwargs)
        return self

    def mixin(self, bases):
        if isinstance(bases, tuple):
            self.__bases__ = bases + self.__bases__
        else:
            self.__bases__ = (bases,) + self.__bases__
        return self

    def __and__(self, other):
        class cls(self, other):
            pass
        return cls

    def __rand__(self, other):
        class cls(other, self):
            pass
        return cls

    def __call__(self, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        obj.params = copy.deepcopy(self.params)
        return obj

    def __matmul__(self, deco):
        return deco(self)


class MetaContainer(ParamType):
    """Meta class of containers

    A container is a algebric system with elements of some type
    and operators acting on the elements

    Example:

        ```
        from collections import UserString
        class C(metaclass=MetaContainer):
            # container of strings
            element_class = UserString
            alias = {'strings': 'elements'}

        c = C(strings=[UserString('I'), UserString('love'), UserString('you')], lasting='for ever')
        print(c.element_class)
        print(c.strings)
        print(c.lasting)
        print(c.n_strings)
        print(c[1])
        for a in c:
            print(a)

        c.regester('upper')
        print(c.upper())
        ```

    Output:

        <class 'collections.UserString'>
        <property object at 0x1065715e0>
        ['I', 'love', 'you']
        for ever
        3
        love
        I
        love
        you
        ['I', 'LOVE', 'YOU']
    """

    def __new__(cls, name, bases, attrs):
        """
        Users have to define `element_class` in the class.
        """

        if 'element_class' in attrs:
            element_class = attrs['element_class']
        else:
            for base in bases:
                if hasattr(base, 'element_class'):
                    element_class = base.element_class
                    break
            else:
                raise Exception('Have not provided element class yet.')
        
        def _getitem(self, k):
            # print(DeprecationWarning('get item directly is not recommended now.'))
            try:
                return self.__elements[k]
            except:
                if isinstance(k, Iterable):
                    return [self[_] for _ in k]
                else:
                    raise TypeError(f'The index must be int/tuple/slice or list-like iterable object! But what you provided is {type(k)}')

        def _setitem(self, k, v):
            # print(DeprecationWarning('get item directly is not recommended now.'))
            self.__elements[k] =v
            self.after_setter()

        def _iter(self):
            return iter(self.__elements)

        def _len(self):
            return len(self.__elements)

        def _contains(self, e):
            return e in self.__elements

        attrs.update(
            {'__getitem__': _getitem,
            '__setitem__': _setitem,
            '__iter__': _iter,
            '__len__': _len,
            '__contains__': _contains}
            )

        def _apply(self, f, *args, **kwargs):
            return self.map(lambda o: f(o, *args, **kwargs), self.__elements)

        @property
        def _n_elements(self):
            return len(self.__elements)

        @property
        def _elements(self):
            return self.__elements

        @_elements.setter
        def _elements(self, x):
            self.__elements = x
            if hasattr(self, 'after_setter'):
                self.after_setter()

        attrs.update(
            {"elements": _elements,
            "n_elements": _n_elements,
            "apply": _apply
            })

        # def _type_check(self):
        #     return all(isinstance(elm, self.element_class) for elm in self.__elements)

        # attrs['type_check'] = _type_check

        def _getstate(self):
            return {'element_class':self.element_class,
            'default_size':self.default_size,
            'elements':self.elements,
            'params':self.params}

        def _setstate(self, state):
            self.element_class = state.get('element_class', self.__class__.element_class)
            self.default_size = state.get('default_size', self.__class__.default_size)
            self.elements = state.get('elements', [])
            self.params = state.get('params', {})

        attrs.update(
            {'__getstate__': _getstate,
            '__setstate__': _setstate
            })

        def _isa(self, cls):
            return isinstance(self, cls) and self.element_class == cls.element_class

        attrs.update({'isa': _isa})

        """
        Regester maps and operands
        if the mapping f is regestered, then A owns method f, automatically
        f(A) := {f(a), a in A} where f is a method of A.
        """

        def _regester_map(self, name, key=None, force=True):
            if key is None:
                key = methodcaller(name)

            _map = self.map if hasattr(self, 'map') else map

            def m(obj, *args, **kwargs):
                return _map(key, obj.__elements, *args, **kwargs)
                
            if not force and hasattr(self, name):
                raise RegesterError(self.__class__, name)
            setattr(self, name, MethodType(m, self))

        attrs.update({
            'regester_map': _regester_map
        })

        def _regester(self, name, key, force=True):
            if not force and hasattr(self, name):
                raise RegesterError(self.__class__, name)
            setattr(self, name, MethodType(key, self))

        attrs.update({
            'regester': _regester
        })

        return super().__new__(cls, name, bases, attrs)

    def __call__(self, *args, **kwargs):
        o = super().__call__()
        o.params = copy.deepcopy(self.params)

        if args:
            elements = []
            if isinstance(self.element_class, tuple):
                for e, e_c in zip(args[0], self.element_class):
                    if not isinstance(e, e_c):
                        try:
                            e = e.copy(type_=e_c)
                        except:
                            e = e_c(e)
                    elements.append(e)
            else:
                e_c = self.element_class
                for e in args[0]:
                    if not isinstance(e, e_c):
                        try:
                            e = e.copy(type_=e_c)
                        except:
                            e = e_c(e)
                    elements.append(e)

            o.__elements = elements
            # for e in o.__elements:  # consider in future
            #     e.__system = o
        else:
            raise TypeError('missing a list/tuple of elements as the unique positional argument!')

        for k, v in kwargs.items():
            setattr(o, k, v)

        return o

    def __getitem__(self, class_):
        """Helper to construct a container

            `C[a] // n` is equiv. to

            ```
            class C(metaclass=MetaContainer):
                element_class = a
                default_size = n
            ```
        
        Args:
            class_: `element_class` of the container
        
        Returns:
            A container class
        """
        return self.set(element_class=class_)

    def __ifloordiv__(self, n):
        return self.set(default_size=n)

    def __floordiv__(self, n):
        class cls(self):
            default_size = n
        return cls

    def random(self, n_elements=None, *args, **kwargs):
        """Generate a container randomly
        
        Arguments:
            **kwargs -- set size of the container if use the alias of n_elements
        
        Keyword Arguments:
            n_elements {number} -- the number of elements (default: {None})
        """

        for k, v in kwargs.items():
            if k in self.alias and self.alias[k] == 'n_elements':
                n_elements = v
                del kwargs[k]
                break
        if isinstance(self.element_class, tuple):
            return self([C.random(*args, **kwargs) for C in self.element_class])
        else:
            n_elements = n_elements or self.default_size
            return self([self.element_class.random(*args, **kwargs) for _ in range(n_elements)])


# class System(MetaContainer):
#     """Metaclass of systems, considered in future!

#     A system is a type of container, that defines operators on them.
#     """
    
#     def __new__(cls, name, bases=(), attrs={}): 
#         """
#         Regester maps and operands
#         if the mapping f is regestered, then A owns method f, automatically
#         f(A) := {f(a), a in A} where f is a method of A.
#         """

#         def _regester_operator(self, name, key=None):
#             if hasattr(self, name):
#                 raise AttributeError(f'`{name}` is an attribute of {self.__class__.__name__}, and would not be regestered.')
#             self.__operators.append(key)
#             setattr(self, name, MethodType(key, self))

#         def _element_regester(self, name, e):
#             if hasattr(self, e):
#                 raise AttributeError(f'`{e}` is an attribute of {self.__class__.__name__}, would not be regestered.')
#             self.__elements.append(e)
#             setattr(self, name, e)

#         attrs.update({
#             '__operators': []
#             'regester_operator': _regester_operator,
#             'regester_element': _regester_element
#         })

#         return super().__new__(cls, name, bases, attrs)

#     def __call__(self, *args, **kwargs):
#         o = super().__call__(*args, **kwargs)
#         for e in o.__elements:  # consider in future
#             e.__system = o
#         return o


class MetaList(MetaContainer):
    # a list is a container of elements with the same type

    def __new__(cls, name, bases, attrs):
        if 'element_class' in attrs:
            element_class = attrs['element_class']
            if isinstance(element_class, tuple):
                raise TypeError('`element_class` should not be a tuple!')
        return super().__new__(cls, name, bases, attrs)

    def __getitem__(self, class_):
        self.element_class = class_
        if isinstance(class_, tuple):
            raise TypeError('`element_class` should not be a tuple!')
        return self

    def __call__(self, *args, **kwargs):
        o = super().__call__(*args, **kwargs)
        for e in o:  # consider in future
            if not isinstance(e, self.element_class):
                raise TypeError(f'"{e}" is not an instance of type "{self.element_class}"')
        return o


class MetaTuple(MetaContainer):
    # a tuple is a container of elements with different types

    def __new__(cls, name, bases, attrs):
        # constructor of MetaMultiContainer
        if 'element_class' in attrs:
            element_class = attrs['element_class']
            if not isinstance(element_class, tuple):
                raise TypeError('`element_class` should be a tuple!')
        return super().__new__(cls, name, bases, attrs)

    def __floordiv__(self, n):
        raise DeprecationWarning('It is meaningless to do `//` on the class {self.__name__} with a number.')

    def __call__(self, *args, **kwargs):
        o = super().__call__(*args, **kwargs)
        for e, t in zip(o, self.element_class):  # consider in future
            if not isinstance(e, t):
                raise TypeError(f'"{e}" is not an instance of type "{t}"')
        return o


class MetaHighContainer(MetaContainer):
    # High order container is a container of containers.

    def __new__(cls, name, bases, attrs):
        if 'element_class' in attrs:
            element_class = attrs['element_class']
            if (not isinstance(element_class, MetaContainer)
                and isinstance(element_class, tuple) and not any(isinstance(ec, MetaContainer) for ec in element_class)
                and not isinstance(element_class, ParamType)):
                raise TypeError('`element_class` should be an instance of `MetaContainer`, or a tuple where one element is an instance of `MetaContainer`.')

        def _flatten(self, type_):
            from toolz import concat
            return concat(elm.__elements for elm in self.__elements)

        attrs['flatten'] = _flatten

        return super().__new__(cls, name, bases, attrs)


class MetaSingle(MetaContainer):
    # metaclass for the containers that have only one element

    def __new__(cls, name, bases, attrs):
        if 'n_elements' in attrs and attrs['n_elements'] !=1:
            raise ValueError('n_elements should be 1!')
        else:
            attrs['n_elements'] = 1
        return super().__new__(cls, name, bases, attrs)

    def __call__(self, *args, **kwargs):
        o = super().__call__(*args, **kwargs)

        if o.n_elements != 1:
            raise ValueError('There should be only 1 element!')
        return o


# import array
# import numpy as np

# def array_check(bases):
#     if array.array in bases or np.ndarray in bases:
#         return True
#     else:
#         for base in bases:
#             if issubclass(base, (array.array, np.ndarray)):
#                 return True
#         else:
#             return False


class MetaArray(ParamType):

    """Metaclass for chromosomes

    Chromosomes could be seen as a container of genes. But we implement them
    by the arrays for convenience.
    """

    def __new__(cls, name, bases, attrs):
        if 'element_class' in attrs:
            element_class = attrs['element_class']
        else:
            for base in bases:
                if hasattr(base, 'element_class'):
                    element_class = base.element_class
                    break
            else:
                raise Exception('Have not provided element class yet.')
        # if not issubclass(element_class, (int, float, np.int_, np.float_, np.bool_)) or isinstance(element_class, str):
        #     raise TypeError('The types of elements should be a subclass of int or float, or a string representing a type')

        # if not array_check(bases):
        #     raise Exception(f'The class `{name}` should be a subclass of numpy.ndarray or array.array!')

        return super().__new__(cls, name, bases, attrs)

    def __ifloordiv__(self, n):
        return self.set(default_size=n)

    def __floordiv__(self, n):
        class cls(self):
            default_size = n
        return cls




#!/usr/bin/env python3

"""GA for linear regression
"""

import numpy as np
from sklearn.cluster import KMeans
from ..learn import BaseEstimator

from .population import StandardPopulation


class GALinearRegression(BaseEstimator, KMeans):
    '''Linear Regression by GA
    '''

    estimated_params = ('cluster_centers_',)

    @classmethod
    def create_model(cls, *args, **kwargs):
        return KMeans(*args, **kwargs)

    @classmethod
    def config(cls, X, Y=None, n_clusters=2, n_individuals=10, *args, **kwargs):

        n_features = X.shape[1]

        class MyIndividual(FloatChromosome // (n_clusters, n_features)):

            def decode(self):
                model = cls.create_model(*args, **kwargs)
                model.cluster_centers_ = np.asarray(self)
                return model

            def _fitness(self):
                model = self.decode()
                return model.score(X)

        MyPopulation = StandardPopulation[MyIndividual]

        pop = MyPopulation.random(n_individuals=n_individuals)

        return pop



#!/usr/bin/env python3

"""
Helper functions
"""

from random import random, randint
from math import exp
from copy import deepcopy

from toolz import unique
import numpy as np
from scipy.spatial.distance import euclidean
from scipy.special import softmax
from scipy.stats import rv_discrete


def copy(obj, *args, **kwargs):
    if hasattr(obj, 'copy'):
        return obj.copy(*args, **kwargs)
    else:
        return deepcopy(obj)


def boltzmann_select(xs, fs, T=1):
    ps = softmax(np.array(fs) /T)
    rv = rv_discrete(values=(np.arange(len(xs)), ps))
    k = rv.rvs()
    return xs[k]


def choice(xs, *args, **kwargs):
    """Choose xi from xs with certain probability
    
    Args:
        xs (List): a list of objects
        ps (List): a list of numbers
        n (int, optional): number of samples
    
    Returns:
        List: the sampling result
    """

    ks = np.random.choice(np.arange(len(xs)), *args, **kwargs)
    return [xs[k] for k in ks]


def choice_uniform(xs, *args, **kwargs):
    """Choose xi from xs with certain probability
    
    Args:
        xs (List): a list of objects
        ps (List): a list of numbers
        n (int, optional): number of samples
    
    Returns:
        List: the sampling result
    """

    ks = np.random.randint(0, len(xs), *args, **kwargs)
    return [xs[k] for k in ks]


def choice_with_fitness(xs, fs=None, n=1, T=1):
    if fs is None:
        fs = [x.fitness for x in xs]
    ps = softmax(np.asarray(fs) / T)
    return choice(xs, p=ps, size=n)


def randint2(lb=0, ub=9, ordered=False):
    """Select two different numbers in [lb, ub] randomly
    
    Formally i != j ~ U(lb, ub)
    Applied in GA operations.
    
    Keyword Arguments:
        lb {number} -- lower bound of interval (default: {0})
        ub {number} -- upper bound of interval (default: {9})
    
    Returns:
        two numbers
    """
    
    i = randint(lb, ub)
    d = ub - lb
    j = randint(i+1, d+i)
    if j > ub:
        j -= (d + 1)
    if ordered:
        if j < i:
            return j, i
    return i, j


def max0(x):
    return np.maximum(x, 0)


def hl(x):
    return np.clip(x, 0, 1) 


def metropolis_rule(D, T, epsilon=0.000001):
    """
    Metropolis rule
    
    Args:
        D (float): number representing the change of the value
        T (float): A number representing temperature
        epsilon (float, optional): The l.b. of T
    
    Returns:
        bool: change the state or not
    """
    if D < 0:
        p = exp(D / max(T, epsilon))
        return random() < p
    else:
        return True


# def proportion(n, N):
#     if n is None:
#         n = N
#     elif 0 < n < 1:
#         n = int(N * n)
#     return n


def pattern(chromosomes):
    """Get the pattern of the chromosomes
    
    Args:
        chromosomes (TYPE): A set of binary chromosomes
    
    Returns:
        str: the pattern of the chromosomes

    Example:
        >> pattern([[0,1,0], [1,0,0]])
        >> **0
    """

    return ''.join([str(a[0]) if all(ai==a[0] for ai in a) else '*' for a in zip(*chromosomes)])


def rotations(x, y):
    """The rotations transforming x to y
    
    Args:
        x, y (array-like): a permutation of objects
    
    Returns:
        list of tuple: each tuple represent a rotation (based on indexes)

    Example:
        >> rotation([5,2,3,1,4], [2,5,3,4,1])
        >> [(0, 1), (3, 4)]
    """
    
    l = []
    for i, xi in enumerate(x):
        yi = y[i]
        if xi != yi:
            for j, t in enumerate(l):
                if x[t[-1]] == xi:
                    if x[t[0]] == yi:
                        break
                    else:
                        l[j] = (*t, x.index(yi))
                        break
                else:
                    if x[t[0]] == yi:
                        l[j] = (j, *t)
                        break
            else:
                l.append((i, x.index(yi)))
    return l


def rotate(x, rotations):
    """Permutate x by the list of rotations `rotation`
    
    Args:
        x (array-like): a permutation
        rotation (list of tuples): a list of rotations
    
    Returns:
        array-like:
    """
    
    if isinstance(rotations, tuple):
        r = rotations
        t = x[r[0]]
        for a, b in zip(r[:-1],r[1:]):
            x[a] = x[b]
        x[r[-1]] = t
    else:
        for r in rotations:
            x = rotate(x, r)
    return x


def prufer_decode(x, nodes=None):
    """Prufer code to tree
    
    Args:
        x (TYPE): Prufer code
        nodes (None, optional): all nodes of the tree
    
    Returns:
        list of pair, representing a tree

    Example:
        >> x = [2,5,6,8,2,5]
        >> print(prufer_decode(x))
        [(1, 2), (3, 5), (4, 6), (6, 8), (7, 2), (2, 5), (5, 8)]
    """
    if nodes is None:
        nodes = np.arange(1, len(x)+3)

    S = set(nodes) - set(x)
    edges = []
    while x:
        i = min(S)
        j = x[0]
        edges.append((i, j))
        S.remove(i)
        x.pop(0)
        if j not in x:
            S.add(j)
        print(x,S)
    edges.append(tuple(S))
    return edges



#!/usr/bin/env python3

"""GA for linear regression
"""

import numpy as np
import numpy.linalg as LA
from sklearn.linear_model import LinearRegression
from ..learn import BaseEstimator

from ..population import StandardPopulation
from digit_converter import IntervalConverter

c = IntervalConverter(lb=-60, ub=60)
class _BinaryChromosome(BinaryChromosome):
    def decode(self):
        return c(self)


class GALinearRegression(BaseEstimator, LinearRegression):
    '''Linear Regression by GA
    '''

    estimated_params = ('coefs_', 'intercepts_')

    @classmethod
    def create_model(cls, *args, **kwargs):
        return LinearRegression(*args, **kwargs)

    @classmethod
    def config(cls, X, Y, n_individuals=10, *args, **kwargs):

        input_dim = X.shape[1]
        assert np.ndim(Y) == 1, 'only support 1D array for `Y`'
        # output_dim = Y.shape[1]

        class MyIndividual(MixedIndividual):

            element_class = FloatChromosome, _BinaryChromosome

            def decode(self):
                model = cls.create_model(*args, **kwargs)
                model.coef_ = np.asarray(self[0])
                model.intercept_ = self[1].decode()
                return model

            def _fitness(self):
                model = self.decode()
                return model.score(X, Y)

        MyPopulation = StandardPopulation[MyIndividual]

        pop = MyPopulation.random(n_individuals=n_individuals, size=(input_dim, 8))

        return pop



#!/usr/bin/env python3


import numpy as np
import numpy.linalg as LA

from ..benchmarks import BaseProblem


class NMF(BaseProblem):
    # M ~ A diag(C) B'
    def __init__(self, M):
        self.M = M
        self.norm = LA.norm(M)

    @staticmethod
    def random(N=500, p=100):
        M = np.random.rand(N, p) * 10
        s = M.sum(axis=1)
        for k in range(N):
            M[k] /= s[k]
        return NMF(M=M)

    def __call__(self, A, B, C=None):
        """A: N * K
        C: K
        B: K * p
        """
        c = A.shape[1]
        if C is not None:
            for i in range(c):
                A[:,i] = A[:,i] * C[i]
        return - LA.norm(self.M - np.dot(A, B)) / self.norm



#!/usr/bin/env python

"""
The Bat Algorithm is a nature-inspired optimization algorithm developed by Xin-She Yang in 2010.
It is based on the echolocation behavior of bats.
Bats emit ultrasonic pulses and listen to the echoes to determine the distance to obstacles and the location of prey.
This behavior forms the basis of the algorithm where solutions are represented as virtual bats,
and their positions in the search space are adjusted iteratively to find the optimal solution. 
"""


from math import exp
from random import random

import numpy as np

from . import BaseIndividual, BasePopulation, FloatChromosome
from .pso import BaseParticle


class Bat(BaseParticle):

    params = {'frequency': 0.5,
        'pulse_rate': 0,
        'loudness': 0,
        'scale': .1
        }

    @property
    def position(self):
        return self.chromosomes[0]

    @position.setter
    def position(self, x):
        self.chromosomes[0] = x
        self.clear_cache()

    @property
    def velocity(self):
        return self.chromosomes[-1]

    @velocity.setter
    def velocity(self, v):
        self.chromosomes[-1] = v

    def move(self):
        self.position += self.velocity * self.scale


class Bats(BasePopulation):

    alias = {'bats': 'elements', 'n_bats': 'n_elements'}

    params = {
        'gamma': 0.5,
        'alpha': 0.95,
        'pulse_rate': 0.9,
        'scaling': 0.2
        }
 
    def init(self):
        for bat in self:
            bat.backup(check=False)
            bat.frequency = random()
            bat.init_pulse_rate = random()*0.5 + 0.5
        super().init()

    def transition(self, k):
        max_frequency = max(self.get_all('frequency'))
        min_frequency = min(self.get_all('frequency'))
        mean_loudness = np.mean(list(self.get_all('loudness')))
        for bat in self:
            bat.frequency = random() * (max_frequency-min_frequency) + min_frequency
            bat.velocity += (self.best_individual.memory['position'] - bat.position) * bat.frequency
            bat.move()

            # local search
            for i, pi in enumerate(bat.position):
                if random() > bat.pulse_rate:
                    r = random()*2 - 1
                    bat.position[i] = self.best_individual.memory['position'][i] + mean_loudness * r * self.scaling

            # update the params
            bat.pulse_rate = bat.init_pulse_rate * (1 - exp(-self.gamma * (k+1)))
            bat.loudness *= self.alpha
            if random() < bat.loudness:
                bat.backup()




#!/usr/bin/env python3

"""
Chromosome classes, subclass of BaseChromosome

A chromosome is an array of genes, or it can be customized by the user.
It could be a part of an individual or encodes a solution directly.
"""

from random import choice, randint, gauss, random

import numpy as np
from scipy.stats import norm
from scipy.special import softmax

from .base import BaseChromosome, BaseGene
from .gene import *
from .utils import *
from .deco import side_effect


def _asarray(out):
    return np.asarray(out) if isinstance(out, np.ndarray) else out


class NumpyArrayChromosome(BaseChromosome, np.ndarray):
    """Chromosome implemented by `np.array`
    
    Attributes:
        element_class (TYPE): the type of gene
    """

    element_class = BaseGene

    def __new__(cls, array=None, element_class=None):
        if element_class is None:
            element_class = cls.element_class
        if array is None:
            array = []

        return np.asarray(array, dtype=element_class).view(cls)

    def __array_finalize__(self, obj):
        if obj is None:
            return
        if isinstance(obj, (tuple, list)):
            obj = self.__class__(obj)
        if isinstance(obj, BaseChromosome):
            self.element_class = getattr(obj, 'element_class', None)

    def __array_ufunc__(self, ufunc, method, *inputs, out=None, **kwargs):

        inputs = map(_asarray, inputs)
        if out is not None:
            out = tuple(map(_asarray, out))

        results = super().__array_ufunc__(ufunc, method, *inputs, out=out, **kwargs)

        if results is NotImplemented:
            return NotImplemented
        elif ufunc.nout == 1:
            if out is None:
                return self.__class__(results) if isinstance(results, np.ndarray) else results
            else:
                out = out[0]
                return self.__class__(out) if isinstance(out, np.ndarray) else out
        else:
            if out is None:
                return tuple(map(lambda o: self.__class__(o) if isinstance(o, np.ndarray) else o, results))
            else:
                return tuple(map(lambda o: self.__class__(o) if isinstance(o, np.ndarray) else o, out))

    @property
    def elements(self):
        return np.asarray(self)

    @elements.setter
    def elements(self, x):
        self.__elements = np.asarray(x)

    @classmethod
    def random(cls, *args, **kwargs):
        if 'size' not in kwargs:
            if cls.default_size:
                kwargs['size'] = cls.default_size
            else:
                raise UnknownSizeError(cls)
        return cls(cls.element_class.random(*args, **kwargs))

    def cross(self, other):
        """Crossover operation for a single chromosome

        Note that when len(self) == 2  ==>  k==1
        
        Args:
            other (BaseChromosome): another chromosome
        
        Returns:
            BaseChromosome: new chromosome, as the child of the two chromosomes
        """

        k = randint(1, len(self)-1)
        return self.__class__(np.concatenate((self[:k], other[k:]), axis=0))

    def cross2(self, other):
        # return 2 children after crossover
        k = randint(1, len(self)-1)
        return (self.__class__(np.concatenate((self[:k], other[k:]), axis=0)),
                self.__class__(np.concatenate((other[:k], self[k:]), axis=0)))

    def copy(self, type_=None, *args, **kwargs):
        type_ = type_ or self.__class__
        return type_(np.copy(self))

    def clone(self):
        """Alias of `copy`, but regarded as a genetic operation
        
        Returns:
            BaseChromosome
        """

        return self.__class__(np.copy(self))

    @side_effect
    def mutate(self, indep_prob=0.1):
        # mutation
        for i in range(len(self)):
            if random() < indep_prob:
                self[i] = self.element_class.random()


class VectorChromosome(NumpyArrayChromosome):
    pass


class MatrixChromosome(NumpyArrayChromosome):
    
    @side_effect
    def mutate(self, indep_prob=0.1):
        r, c = self.shape
        for i in range(r):
            for j in range(c):
                if random() < indep_prob:
                    self[i, j] += gauss(0, 0.1)

    def cross(self, other):
        r, c = self.shape
        k, l = randint(1, r-1), randint(1, c-1)
        A = np.vstack((self[:k, :l], other[k:, :l]))
        B = np.vstack((other[:k, l:], self[k:, l:]))
        return self.__class__(np.hstack((A, B)))


class NaturalChromosome(VectorChromosome):

    element_class = NaturalGene

    @side_effect
    def mutate(self, indep_prob=0.1):
        for i in range(len(self)):
            if random()< indep_prob:
                self[i] = NaturalGene.random()

    def dual(self):
        return self.__class__(self.element_class.ub - self)


class DigitChromosome(NaturalChromosome):

    element_class = DigitGene

    def __str__(self):
        return "".join(map(str, self))


class BinaryChromosome(NaturalChromosome):

    element_class = BinaryGene

    def __str__(self):
        return "".join(map(str, self))

    @classmethod
    def zero(cls):
        return cls(np.zeros(cls.default_size))

    @classmethod
    def one(cls):
        return cls(np.ones(cls.default_size))

    @side_effect
    def mutate(self, indep_prob=0.5):
        for i in range(len(self)):
            if random() < indep_prob:
                self[i] ^= 1

    def dual(self):
        return self.__class__(1 ^ self)


class PermutationChromosome(NaturalChromosome):
    # A chromosome representing a permutation

    element_class = NaturalGene
    default_size = 10

    @classmethod
    def identity(cls):
        return cls(np.arange(cls.default_size))

    @classmethod
    def random(cls, size=None):
        size = size or cls.default_size
        return cls(np.random.permutation(cls.default_size))

    # def __sub__(self, other):
    #     return rotations(self, other)

    def move_toward(self, other):
        r = choice(rotations(self, other))
        rotate(self, r)

    @side_effect
    def mutate(self):
        i, j = randint2(0, self.default_size-1)
        self[[i,j]] = self[[j,i]]

    def cross(self, other):
        k = randint(1, len(self)-2)
        return self.__class__(np.hstack((self[:k], [g for g in other if g not in self[:k]])))

    def __str__(self):
        if len(self)>10:
            return ",".join(map(str, self))
        return "".join(map(str, self))

    def dual(self):
        return self[::-1]


class FloatChromosome(NumpyArrayChromosome):

    element_class = FloatGene
    sigma = 0.1

    def __str__(self):
        return "|".join(format(c, '.4') for c in self)

    @side_effect
    def mutate(self, indep_prob=0.1, mu=0, sigma=None):
        sigma = sigma or self.sigma
        for i in range(len(self)):
            if random() < indep_prob:
                self[i] += gauss(mu, sigma)

    def random_neighbour(self, mu=0, sigma=None):
        # select a neighour randomly
        sigma = sigma or self.sigma
        cpy = self.copy()
        n = norm(mu, sigma)
        cpy += n.rvs(len(self))
        return cpy


class FloatMatrixChromosome(MatrixChromosome, FloatChromosome):
    pass


class PositiveChromosome(FloatChromosome):

    def normalize(self):
        self[:] = max0(self)


class UnitFloatChromosome(PositiveChromosome):

    element_class = UnitFloatGene

    def dual(self):
        return UnitFloatChromosome(1 - self)

    def tobinary(self):
        return self >= 0.5

    def mutate(self, *args, **kwargs):
        super().mutate(*args, **kwargs)
        self.normalize()

    def normalize(self):
        self[:] = hl(self)


class ProbabilityChromosome(PositiveChromosome):
    """ProbabilityChromosome

    The genes represent a distribution, such as [0.1,0.2,0.3,0.4].
    
    Extends:
        PositiveChromosome
    """

    element_class = UnitFloatGene

    @classmethod
    def random(cls, size=None):
        if size is None:
            if cls.default_size:
                size = cls.default_size
            else:
                raise UnknownSizeError(cls)
        return cls(np.random.dirichlet(np.ones(size)))

    def check(self):
        assert np.sum(self) == 1, 'the sum of the chromosome must be 1!'

    # def mutate(self, indep_prob=0.1):
    #     """Mutation of ProbabilityChromosome
    #     if mutation happend on two genes i and j, then select a number r randomly
    #     i <- r, j <= p - r, where p is the sum of the original probs of i and j.
        
    #     Keyword Arguments:
    #         indep_prob {number} -- independent prob of mutation for any gene (default: {0.1})
        
    #     Returns:
    #         ProbabilityChromosome -- new obj after mutation
    #     """
    #     for i in range(len(self)-1):
    #         if random() < indep_prob:
    #             j = randint(i+1, len(self)-1)
    #             p = self[i] + self[j]
    #             r = np.random.uniform(0, p)
    #             self[i] = r
    #             self[j] = p-r
    #     return self

    def cross(self, other):
        k = randint(1, len(self)-2)
        array = np.hstack((self[:k], other[k:])) + 0.001
        array /= array.sum()
        return self.__class__(array)

    def random_neighbour(self):
        # select a neighour randomly
        cpy = self.copy()
        i, j = randint2(0, len(cpy)-1)
        p = cpy[i] + cpy[j]
        r = random() * p
        cpy[[i,j]] = [r, p - r]
        return cpy

    def mutate(self, *args, **kwargs):
        super().mutate(*args, **kwargs)
        self.normalize()

    def normalize(self):
        self[:] = softmax(self)


class StochasticMatrixChromosome(MatrixChromosome, PositiveChromosome):

    def normalize(self):
        self[:] = softmax(self, axis=1)

class CircleChromosome(FloatChromosome):
    """Used in Quantum-Chromosome
    
    Extends:
        FloatChromosome
    """

    element_class = CircleGene

    def mutate(self, *args, **kwargs):
        super().mutate(*args, **kwargs)
        self.normalize()

    def normalize(self):
        self[:] = self % self.element_class.period


class QuantumChromosome(CircleChromosome):
 
    _measure_result = None

    @property
    def measure_result(self):
        if self._measure_result is None:
            self.measure()
        return self._measure_result

    def measure(self):
        # measure a quantum chromosome to get a binary sequence
        rs = np.random.random(size=(len(self),))
        self._measure_result = np.asarray(np.cos(self) ** 2 > rs, dtype=np.int_)

    def decode(self):
        return self.measure_result


# Implement Chromosome class by `array.array`

import copy
import array

class ArrayChromosome(BaseChromosome, array.array):
    """Chromosome class implemented by `array.array`
    
    Attributes:
        element_class (str): the type of gene
    """

    element_class = 'd'

    def __new__(cls, array=None, element_class=None):
        if element_class is None:
            element_class = cls.element_class
        if array is None:
            array = []

        return array.array(element_class, array)

    def cross(self, other):
        # note that when len(self) == 2  ==>  k==1
        k = randint(1, len(self)-1)
        return self[:k] + other[k:]

    def copy(self, type_=None):
        return copy.copy(self)

    @side_effect
    def mutate(self, indep_prob=0.1):
        a = self.random()
        for k, ak in enumerate(a):
            if random() < indep_prob:
                self[k] = ak


class ListChromosome(BaseChromosome, list):
    """Chromosome class implemented by `list`
    
    Attributes:
        element_class (str): the type of gene
    """

    element_class = None

    def cross(self, other):
        # note that when len(self) == 2  ==>  k==1
        k = randint(1, len(self)-1)
        return self[:k] + other[k:]

    def copy(self, type_=None):
        return copy.deepcopy(self)

    @side_effect
    def mutate(self, indep_prob=0.1):
        a = self.random()
        for k, ak in enumerate(a):
            if random() < indep_prob:
                self[k] = ak



#!/usr/bin/env python3

from typing import Iterable


def _call(s, obj):
    """Core function of `ezstat`

    A functional extension for s(obj) or obj.s or obj.s()
    If s is an string, then it only returns obj.s or obj.s().

    s could be a tuple or dict, as a lifted form, though it is not recommended.
    
    Arguments:
        s {function | string} -- Statistics
        obj -- object of statistics
    
    Return:
        a number or a tuple of numbers, as the value of statistics
    """
    if isinstance(s, str):
        if not hasattr(obj, s):
            raise ValueError(f"the object '{obj}' of '{obj.__class__}' has no attribute '{s}'")
        f = getattr(obj, s)
        r = f() if callable(f) else f
    elif callable(s):
        r = s(obj)
    elif isinstance(s, (int, float)):
        print(Warning('Deprecated to use a constant number!'))
        r = s
    elif isinstance(s, tuple):
        return tuple(_call(si, obj) for si in s)
    elif isinstance(s, dict):
        return {i:_call(si, obj) for i, si in s.items()}
    else:
        raise TypeError(f"The type of `{s}` is not permissible!") 

    return r


class Statistics(dict):
    """
    Statistics is a type of dict{str:function},
    where `function` will act on the object of statistics.

    As the value of dict, `function` has not to be a function.
    If it is a string, then the attribute of object will be called.
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(cls, *args, **kwargs)
        for k, _ in obj.items():
            if not isinstance(k, str):
                raise TypeError(f'The keys must be strings, but `{k}` is not a string.')
        return obj

    def __call__(self, obj, split:bool=False):

        res = {}
        for k, s in self.items():
            if s is True:
                s = k.lower().replace(' ', '_').replace('-', '_')
            r = _call(s, obj)
            if split and isinstance(r, Iterable):
                res.update({f"{k}[{i}]": ri for i, ri in enumerate(r)})
            elif isinstance(r, dict):
                res.update({f"{k}[{i}]": ri for i, ri in r.items()})
            else:
                res[k] = r
        return res




#!/usr/bin/env python3

"""Slime Mould Algorithm
"""

from scipy.spatial.distance import pdist, squareform
import numpy as np

from .base import BaseIndividual
from .mixin import PopulationMixin, FitnessMixin
from .chromosome import FloatChromosome
from .population import HOFPopulation
from random import random

from .deco import basic_memory
from .utils import randint2


@basic_memory
class SlimyMaterial(FitnessMixin):

    def approach_food(self, fame, direction, p, vb, vc):
        if random() < p:
            self[:] = (fame + vb * direction)[:]
        else:
            self[:] = (vc * self)[:]

    def random_move(self):
        raise NotImplementedError


class SlimeMould(HOFPopulation):
    """Slime Mould Algorithm
    """

    element_class = SlimyMaterial

    params = {
        "n_iter": 100
    }

    def get_ranks(self):
        all_fitness = np.array([i._fitness() for i in self])
        max_fitness = np.max(all_fitness)
        min_fitness = np.min(all_fitness)
        return (max_fitness - all_fitness)/(max_fitness - min_fitness)

    def approach_food(self, t):
        N = len(self)
        # calculate vc and a
        vc = 1 - t/self.n_iter
        a = np.arctanh(vc)
        # all fitness, max/min fitness
        all_fitness = self.get_all_fitness()
        max_fitness = np.max(all_fitness)
        # calculate p and w
        ps = np.tanh(np.subtract(max_fitness, all_fitness))
        ks = np.argsort(all_fitness)
        sign = np.ones(N); sign[ks[:N//2]] = -1
        ws = 1 + sign * random() * np.log1p(self.get_ranks())
        fame = self[ks[-1]]
        for sm, p, w in zip(self, ps, ws):
            i, j = randint2(0, N-1)
            direction = w * self[i] - self[j]
            vb = a * (2*random() - 1)
            if random()< 0.03:
                sm.random_move()
            else:
                sm.approach_food(fame, direction, p, vb, vc)
    
    def transition(self, t):

        self.approach_food(t)
        for i in self:
            i.backup()



#!/usr/bin/env python3

"""
The Gravity Searching Algorithm (GSA) is a metaheuristic optimization method that simulates the law of gravity in physics.
It was introduced as a nature-inspired algorithm for solving optimization problems, particularly in continuous domains.

*References*
1. Rashedi, E., Nezamabadi-Pour, H., & Saryazdi, S. (2009). "GSA: A Gravitational Search Algorithm". Information Sciences, 179(13), 2232-2248.
2. Rashedi, E., Nezamabadi-Pour, H., & Saryazdi, S. (2011). "A New Method for Solving Optimization Problems Using Gravitational Search Algorithm". International Journal of Computer Applications, 22(8), 1-6.
3. Niazi, M., Mirjalili, S., Mirjalili, S. M., & Yang, X. S. (2016). "Enhanced Gravity Search Algorithm". Swarm and Evolutionary Computation, 6(1), 10-21.
"""

from scipy.spatial.distance import pdist, squareform
from .base import PopulationMixin
from .chromosome import FloatChromosome
from .pso import BaseParticle
from .utils import euclidean, random, exp, metropolis_rule
from .deco import side_effect

import numpy as np


class Particle(BaseParticle):
    """A particle in GSA
    
    Extends:
        PolyIndividual
    
    Variables:
        default_size {number} -- one individual represented by 2 chromosomes: position and velocity
        phantom {Particle} -- the current state of the particle moving in the solution space.
    """

    element_class = FloatChromosome
    default_size = 2

    params = {'accelerate': 0}

    @property
    def position(self):
        return self.chromosomes[0]

    @position.setter
    def position(self, x):
        self.chromosomes[0] = x
        self.after_setter()

    @property
    def velocity(self):
        return self.chromosomes[1]

    @velocity.setter
    def velocity(self, v):
        self.chromosomes[1] = v

    @side_effect
    def move(self):
        """Moving the particl with Newton's mechanics
        """
        r = random()
        cpy = self.copy(fitness=None)
        cpy.velocity = r * cpy.velocity + cpy.accelerate
        cpy.position = cpy.position + cpy.velocity
        flag = metropolis_rule(D=cpy.fitness - self.fitness, T=10)
        D = cpy.fitness - self.fitness
        if flag:
            self.chromosomes = cpy.chromosomes


class GravitySearch(PopulationMixin):
    """Standard GSA
    
    Extends:
        PopulationMixin
    """

    alias = {'particles': 'elements',
    'n_particles': 'n_elements'}

    element_class = Particle
    default_size = 20

    params = {'gravity_coefficient': 100, 'attenuation_coefficient': 10}

    def compute_mass(self):
        fitnesses = np.asarray([particle.fitness for particle in self])
        worst_fitness = np.min(fitnesses)
        best_fitness = np.max(fitnesses)
        epsilon = 0.0001
        m = (fitnesses - worst_fitness + epsilon) / (best_fitness - worst_fitness + epsilon)
        return m / m.sum()

    def compute_accelerate(self):
        # compute force
        D = np.array([[pj.position - pi.position for pi in self] for pj in self])
        R = squareform(pdist([p.position for p in self]))
        for i in range(self.n_particles):
            R[i, i]=1
        m = self.compute_mass()
        M = np.tile(m, (self.n_particles, 1))
        for i in range(self.n_particles):
            M[i, i]=0
        M /= R**3
        M *= self.gravity_coefficient * np.random.random((self.n_particles, self.n_particles))
        A = M[:,:,None] * D
        A = A.sum(axis=0)

        # set accelerate
        for i, particle in enumerate(self):
            particle.accelerate = A[i, :]

    def transition(self, k):
        """
        Transitation of the states of particles
        """
        self.compute_accelerate()
        self.move()
        self.gravity_coefficient = exp(-self.attenuation_coefficient*k / self.n_iter)

    def move(self):
        for particle in self:
            particle.move()




#!/usr/bin/env python3


from scipy.stats import norm

from ..mixin import FitnessMixin


class RandomWalk(FitnessMixin):
    """Random Walk Algo.
    """

    params = {'sigma': 1}

    def transit(self, k, *args, **kwargs):
        sigma *= self.sigma * 0.99**k
        cpy = self.copy(fitness=None)
        cpy.mutate(sigma)

        D = cpy.fitness - self.fitness
        if D > 0:
            self.chromosomes = cpy.chromosomes
            self.fitness = cpy.fitness

    def mutate(self, sigma):
        n = norm(0, sigma)
        self.chromosomes = [chromosome + n.rvs(chromosome.n_genes) for chromosome in cpy]



#!/usr/bin/env python3


class UnknownSizeError(Exception):

    def __init__(self, cls):
        self.cls = cls

    def __str__(self):
        return f'The size of class `{self.cls.__name__}` is unkown, the object could not be generated.'


class UnavalibleAttributeError(AttributeError):

    def __init__(self, cls, attr_name):
        self.cls = cls
        self.attr_name = attr_name

    def __str__(self):
        return f'Did not define attribute `{self.attr_name}` for the class `{self.cls.__name__}`.'


class RegesterError(AttributeError):

    def __init__(self, cls, attr_name):
        self.cls = cls
        self.attr_name = attr_name

    def __str__(self):
        return AttributeError(f'`{attr_name}` is an attribute of the class `{self.cls.__name__}`, and would not be redefined.')



#!/usr/bin/env python3

"""Firefly Algorithm

*References*
Yang, X. S. (2010). "Firefly Algorithm: A New Approach for Optimization".
  In Stochastic Algorithms: Foundations and Applications (pp. 169-178). Springer Berlin Heidelberg.
Yang, X. S. (2013). "Nature-Inspired Metaheuristic Algorithms". Luniver Press.
"""

from random import gauss, random
from scipy.spatial.distance import pdist, squareform
import numpy as np

from .base import PopulationMixin
from .chromosome import FloatChromosome
from .individual import PolyIndividual

from .pso import BaseParticle


@basic_memory
class BaseFirefly(BaseParticle):

    params = {
        'gamma': 1,
        'alpha': 1
    }

    def update_vilocity(self, fame=None, *args, **kwargs):
        raise NotImplementedError

    def random_move(self):
        raise NotImplementedError


def attractiveness(distance, gamma=1.0):
    return np.exp(-gamma * distance**2)


class StandardFireflies:
    """Starndard Firefly Algorithm
    """

    element_class = BaseFirefly

    params = {
        "gamma": 1,
        "beta": 1,
        "alpha": 0.2
    }
    
    def transition(self, *args, **kwargs):

        distances = squareform(pdist([f.position for f in self]))
        att = attractiveness(distances)

        for i, fi in enumerate(self[:-1]):
            for j, fj in enumerate(self[i+1:]):
                if fj.fitness > fi.fitness:
                    fi.move(self.alpha * att[i, j] * (fj.position - fi.position))

        for f in self:
            f.random_move()

        for f in self:
            f.backup()



#!/usr/bin/env python3

"""
Gene classes
"""

import numpy as np
from . import BaseGene


class NaturalGene(np.int_, BaseGene):
    lb, ub = 0, 10

    @classmethod
    def random(cls, *args, **kwargs):
        return np.random.randint(cls.ub, dtype=cls, *args, **kwargs)


class DigitGene(NaturalGene):
    pass


class IntegerGene(np.int_, BaseGene):
    lb, ub = -10, 10

    @classmethod
    def random(cls, *args, **kwargs):
        return np.random.randint(cls.ub-cls.lb, dtype=cls, *args, **kwargs) + cls.lb


class BinaryGene(np.int_, BaseGene):

    @classmethod
    def random(cls, *args, **kwargs):
        return np.random.randint(2, dtype=cls, *args, **kwargs)


class FloatGene(np.float_, BaseGene):
    lb, ub = 0, 1

    @classmethod
    def random(cls, *args, **kwargs):
        if 'size' in kwargs:
            return np.random.uniform(cls.lb, cls.ub, *args, **kwargs).astype(cls, copy=False)
        else:
            return np.random.uniform(cls.lb, cls.ub, *args, **kwargs)


class UnitFloatGene(FloatGene):
    lb, ub = 0, 1


class PeriodicGene(FloatGene):

    @property
    def period(self):
        return ub - lb


class CircleGene(PeriodicGene):
    lb, ub = 0, 2*np.pi
    period = 2*np.pi

    
class SemiCircleGene(CircleGene):
    lb, ub = 0, np.pi



#!/usr/bin/env python3


import numpy as np
import numpy.linalg as LA

from ..benchmarks import BaseProblem

_basis = [lambda x: np.ones(len(x)), lambda x: x, lambda x: x**2, lambda x: x**3, lambda x: x**4, 
np.sin, np.cos, np.tan, np.abs, lambda x:x>0, lambda x:x<0,
np.exp, lambda x: np.log(np.abs(x)+0.1)]

n_basis_ = len(_basis)


class Fitting(BaseProblem):
    def __init__(self, X, y):
        self.X = X
        self.y = y
        self.n = LA.norm(y)

    def random(N, p):
        pass

    def fit(self, *params):
        return np.sum([c*np.tanh(b * self.X - a) for a, b, c in zip(*params)], axis=0)

    def __call__(self, *params):
        yy = self.fit(*params)
        return - LA.norm(self.y-yy, 1) / self.n


class CurveFitting(Fitting):

    def fit(self, *params):
        return np.vstack((np.sum([c*relu(b * self.X - a) for a, b, c in zip(*params[:3])], axis=0), 
            np.sum([c*relu(b * self.X - a) for a, b, c in zip(*params[3:])], axis=0)))


_indicator = lambda x, y: (0<= x) * (x<= 5) * (0<= y) * (y <= 5)

from math import cos, sin

def basis(a, d1, d2, t):
    """basis
    
    Args:
        a (number): Rotation
        d1 (number): Scaling
        d2 (number): Scaling
        t (number): Translation
    
    Returns:
        function
    """

    c = cos(3.14*a)
    s = sin(3.14*a)
    def _f(x,y):
        y, x = np.meshgrid(y-t[0], x-t[1])
        return _indicator(d1*(c*x-s*y), d2*(s*x+c*y))
    return _f


from PIL import Image

class Painting(Fitting):

    def __init__(self, image, size=None, mode=None, channel=0):
        self.size = size or image.size
        self.mode = mode or image.mode
        self.y = np.asarray(image.resize(self.size), np.float_)[:,:,channel]
        self.n = 256

    def fit(self, *params):
        ks = np.arange(self.size[0])
        ls = np.arange(self.size[1])
        return np.sum([c * basis(a, d1, d2, t)(ks, ls) for a, d1, d2, t, c in zip(*params)], axis=0)

    def toimage(self, *params):
        yy = self.fit(*params)
        return Image.fromarray(yy.astype('uint8')).convert(self.mode)




#!/usr/bin/env python

"""
Differential Evolution Algorithm
"""


import numpy as np

from .base import BaseIndividual
from .mixin import PopulationMixin
from .meta import MetaContainer
from .individual import MonoIndividual

from .utils import *


class DifferentialEvolution(PopulationMixin, metaclass=MetaContainer):
    # Differential Evolution Algo.
    
    element_class = BaseIndividual
    default_size = 4

    params ={
        "factor" : 0.05,
        "cross_prob": 0.75,
    }

    def init(self):
        self.ndims = tuple(map(len, self[0]))

    def transition(self, *args, **kwargs):
        self.move()
        for k, (test_individual, individual) in enumerate(zip(self.test, self)):
            if test_individual.fitness > individual.fitness:
                self[k] = test_individual

    def move(self):
        self.test = self.copy()
        for t in self.test:
            x0, x1, x2 = choice(self, size=3, replace=False)
            jrand = map(np.random.randint, self.ndims)
            xx = x0 + self.factor * (x1 - x2)
            for chromosome, xc, n, r in zip(t, xx, self.ndims, jrand):
                ind = np.random.random(n) < self.cross_prob
                chromosome[ind] = xc[ind]
                chromosome[r] = xc[r]


class DifferentialEvolutionC(DifferentialEvolution):
    # For the population of chromosomes

    params = {
        "factor": 0.5,
        "cross_prob": 0.75,
    }

    def init(self):
        self.ndim = len(self[0])

    def move(self):
        self.test = self.copy()
        for t in self.test:
            x0, x1, x2 = choice(self, size=3, replace=False)
            xx = x0 + self.factor * (x1 - x2)
            jrand = np.random.randint(self.ndim)
            ind = np.random.random(n) < self.cross_prob
            t[ind] = xx[ind]
            t[jrand] = xx[jrand]




#!/usr/bin/env python3

"""
The base classes are defined here, mainly for implementation of GAs.

This is the core module of pyrimidine. 

Main classes:

    BaseGene: the gene of chromosome
    BaseChromosome: sequence of genes, represents part of a solution
    BaseIndividual: sequence of chromosomes, represents a solution of a problem
    BasePopulation: set of individuals, represents a set of a problem
                    also the state of a stachostic process
    BaseMultiPopulation/BaseCommunity: set of populations for more complicated optimalization
    BaseEnviorenment:

Remark:
1. Subclass the classes and override some main method esp. `_fitness`.
2. `BaseGene` is not important, as a wrapper of np.int_ and np.float_
3. The base classes have been crafted specifically for GA-style algorithms.
  If your novel algorithm differs from GAs, it is advisable to derive from the mixin classes. 

Example:

    select ti, ni from arraies t, n
    sum of ni ~ 10 (for example), while ti are exptected to be not repeated

    The opt. problem is
    min sum of {ni} and maximum of frequences in {ti}
    where i are selected indexes.

    ```
    t = np.random.randint(1, 5, 100)
    n = np.random.randint(1, 4, 100)

    import collections
    def max_repeat(x):
        # maximum of numbers of repeats
        c = collections.Counter(x)
        bm=np.argmax([b for a, b in c.items()])
        return list(c.keys())[bm]

    class MyIndividual(MonoIndividual):

        element_class = BinaryChromosome

        def _fitness(self):
            x = self.evaluate()
            return - x[0] - x[1]

        def evaluate(self):
            return np.dot(n, self.chromosomes[0]), max_repeat(ti for ti, c in zip(t, self.chromosomes[0]) if c==1)

    class MyPopulation(StandardPopulation):
        element_class = MyIndividual

    pop = MyPopulation.random(n_individuals=50, size=100)
    pop.evolve()
    print(pop.best_individual)
    ```
"""

from operator import attrgetter
import typing
from random import randint, random

from toolz import concat
from itertools import product
import numpy as np

from .errors import *
from .meta import *
from .mixin import *

from .deco import side_effect


class BaseGene:
    """Base class of genes
    
    Attributes:
        values (tuple of numbers): the values that a gene takes
    """
    
    values = (0, 1)

    def __repr__(self):
        return self.__class__.__name__ + f': {self}'

    @classmethod
    def random(cls, *args, **kwargs):
        return cls(np.random.choice(cls.values, *args, **kwargs))


class BaseChromosome(FitnessMixin, metaclass=MetaArray):
    """Base class of chromosomes

    Chromosome is an array of genes. It is the unit of the GA.

    Attributes:
        default_size (int): the default number of genes in the chromosome
        element_class (TYPE): the type of gene
    """
    
    element_class = BaseGene
    default_size = 8

    alias = {
    "chromosomes": "elements",
    "n_chromosomes": "n_elements"
    }

    def __repr__(self):
        return f'{self.__class__.__name__}: {":".join(map(repr, self))}'

    def __str__(self):
        return "|".join(map(str, self))

    def transition(self, *args, **kwargs):
        self.mutate()

    def x(self, other):
        # alias for cross
        return self.cross(other)

    def cross(self, other):
        # crossover operation
        raise NotImplementedError

    @side_effect
    def mutate(self):
        # mutation operation
        raise NotImplementedError

    def replicate(self):
        # Replication operation of a chromosome
        ind = self.copy()
        ind.mutate()
        return ind

    def decode(self):
        """Decoding of the chromesome

        Translate the chromesome to (part of) solution, maybe a number.
        """
        return self

    def equal_to(self, other):
        return np.array_equal(self, other)

    @classmethod
    def random(cls, *args, **kwargs):
        raise NotImplementedError


class BaseIndividual(FitnessMixin, metaclass=MetaContainer):
    """Base class of individuals

    It is essentially a sequence of chromosomes that may vary in sizes.

    You should implement the methods: cross, mute ...
    """

    element_class = BaseChromosome
    default_size = 2

    alias = {"chromosomes": "elements",
    "n_chromosomes": "n_elements"}

    def __repr__(self):
        # seperate the chromosomes with $ 
        sep = " $ "
        return f'{self.__class__.__name__}:= {sep.join(map(repr, self))}'

    def __str__(self):
        sep = " $ "
        return sep.join(map(str, self))

    def __format__(self, spec=None):
        """ 
        Keyword Arguments:
            spec {str} -- set to be `decode` if need decoding (default: {None})
        
        Returns:
            str
        """
        if spec is None:
            return str(self)
        elif spec in {'d', 'decode'}:
            return ' | '.join(map(str, self.decode()))
        else:
            return str(self)

    def _fitness(self):
        if hasattr(self, 'environment'):
            return self.environment.evaluate(self)
        else:
            raise NotImplementedError

    def transition(self, *args, **kwargs):
        self.mutate()

    def cross(self, other):
        # Cross operation of two individual
        return self.__class__([chromosome.cross(other_c) for chromosome, other_c in zip(self, other)])

    def cross2(self, other):
        return self.__class__(list(product(*(chromosome.cross2(other_c) for chromosome, other_c in zip(self, other)))))

    @side_effect
    def mutate(self):
        # Mutating operation of an individual
        for chromosome in self:
            chromosome.mutate()
        return self

    def replicate(self):
        # Replication operation of an individual
        ind = self.copy()
        ind.mutate()
        return ind

    def get_neighbour(self):
        """to get a neighbour of an individual

        e.g. mutate only one gene
        """
        raise NotImplementedError

    def decode(self):
        """Decode an individual to a real solution

        For example, transform a 0-1 sequence to a real number.
        """
        return [chromosome.decode() for chromosome in self]

    def __eq__(self, other):
        return np.all(list(map(np.equal, self, other)))

    def __mul__(self, n):
        """population = individual * n
        
        Args:
            n (TYPE): positive integer
        
        Returns:
            TYPE: BasePopulation
        """
        
        assert isinstance(n, np.int_) and n > 0, 'n must be a positive integer'
        C = StandardPopulation[self.__class__]
        return C([self.copy() for _ in range(n)])

    def __add__(self, other):
        # algebraic operation `+` of two individuals
        return self.__class__([this + that for this, that in zip(self, other)])

    def __sub__(self, other):
        return self.__class__([this - that for this, that in zip(self, other)])

    def __rmul__(self, other):
        # assert isinstance(other, np.number)
        return self.__class__([other * this for this in self.chromosomes])


class BasePopulation(PopulationMixin, metaclass=MetaContainer):
    """The base class of population in GA
    
    Represents a state of a stachostic process (Markov process)
    
    Extends:
        PopulationMixin
    """

    element_class = BaseIndividual
    default_size = 20

    params = {'mate_prob':0.75, 'mutate_prob':0.2, 'tourn_size':5}

    alias = {"individuals": "elements",
        "n_individuals": "n_elements",
        "best_individual": "best_element",
        "worst_individual": "worst_element",
        "best_individuals": "best_elements",
        "get_best_individual": "get_best_element",
        "get_best_individuals": "get_best_elements"
    }

    def __str__(self):
        return '&\n'.join(map(str, self))

    def transition(self, *args, **kwargs):
        """Transitation of the states of population

        It is considered to be the standard flow of the Genetic Algorithm
        """

        self.select()
        self.mate()
        self.mutate()

    def migrate(self, other):
        """Migration from one population to another

        Applied in Multi population GA, 
        where the best individual of one populaitonemigrates to another.
        """
        raise NotImplementedError

    def select(self, n_sel=None, tourn_size=None):
        """The standard method of selecting operation in GA
        
        Select the best individual among `tourn_size` randomly chosen
        individuals, `n_sel` times.
        """

        if n_sel is None:
            n_sel = self.default_size
        elif 0 < n_sel < 1:
            n_sel = int(self.n_individuals * n_sel)
        if n_sel >= self.n_individuals:
            return
        winners = []
        rest = list(range(self.n_individuals))
        size = tourn_size or self.tourn_size
        n_rest = self.n_individuals

        for i in range(n_sel):
            if n_rest == 0:
                break
            elif n_rest <= size:
                aspirants = rest
            else:
                aspirants = np.random.choice(rest, size, replace=False)
            _winner = np.argmax([self[k].fitness for k in aspirants])
            winner = aspirants[_winner]
            winners.append(winner)
            rest.remove(winner)
            n_rest -= 1
        if winners:
            self.individuals = [self[k] for k in winners]
        else:
            raise Exception('No winners in the selection!')

    def merge(self, other, n_sel=None):
        """Merge two populations.

        Applied in the case when merging the offspring to the original population.

        `other` should be a population or a list/tuple of individuals
        """

        if isinstance(other, BasePopulation):
            self.extend(other.individuals)
        elif isinstance(other, typing.Iterable):
            self.extend(other)
        else:
            raise TypeError("`other` should be a population or a list/tuple of individuals")

        if n_sel:
            self.select(n_sel)

    @side_effect
    def mutate(self, mutate_prob=None):
        """Mutate the whole population.

        Just call the method `mutate` of each individual
        
        Keyword Arguments:
            mutate_prob {number} -- the proba. of mutation of one individual (default: {None})
        """

        for individual in self:
            if random() < (mutate_prob or self.mutate_prob):
                individual.mutate()

    def mate(self, mate_prob=None):
        """To mate the entire population.

        Just call the method `mate` of each individual (customizing anthor individual)
        
        Keyword Arguments:
            mate_prob {number} -- the proba. of mating of two individuals (default: {None})
        """
        
        mate_prob = mate_prob or self.mate_prob
        offspring = [individual.cross(other_individual) for individual, other_individual in zip(self[:-1], self[1:])
        if random() < mate_prob]
        self.extend(offspring)
        return offspring

    def mate_with(self, other, mate_prob=None):
        mate_prob = mate_prob or self.mate_prob
        offspring = [individual.cross(other_individual) for individual, other_individual in product(self, other)
        if random() < mate_prob]
        self.extend(offspring)
        return offspring

    @side_effect
    def local_search(self, n_iter=2, *args, **kwargs):
        """Call local searching method
        
        By default, it calls the `ezolve` methods of individuals, iteratively
        """

        for individual in self:
            individual.ezolve(n_iter=n_iter or self.n_iter, init=False)

    def get_rank(self, individual):
        """Get rank of one individual

        Use `rank` if you call it frequently.
        """

        r = 0
        for ind in self.sorted_():
            if ind.fitness <= individual.fitness:
                r += 1
            else:
                break
        individual.ranking = r / self.n_individuals
        return individual.ranking

    def rank(self, tied=False):
        """To rank all individuals by the fitness increasingly
        
        Args:
            tied (bool, optional): for tied ranking
        """
        
        sorted_list = self.sorted_()
        if tied:
            k = 0
            while k < self.n_individuals:
                r = 0
                individual = sorted_list[k]
                for i in sorted_list[k+1:]:
                    if i.fitness == individual.fitness:
                        r += 1
                    else:
                        break
                for i in sorted_list[k:k+r+1]:
                    i.ranking = (r + k) / self.n_individuals
                k += r + 1
        else:
            for k, i in enumerate(sorted_list):
                i.ranking = k / self.n_individuals

    def cross(self, other):
        # Cross two populations as two individuals
        k = randint(1, self.n_individuals-2)
        return self.__class__(self[k:] + other[:k])

    def migrate(self, other):
        # migrate between two populations
        k = randint(1, self.n_individuals-2)
        self.individuals = self[k:] + other[:k]
        other.individuals = other[k:] + self[:k]

    def dual(self):
        return self.__class__([c.dual() for c in self])


class BaseMultiPopulation(PopulationMixin, metaclass=MetaHighContainer):
    """Base class of BaseMultiPopulation
    
    Attributes:
        default_size (int): the number of populations
        element_class (BasePopulation): the type of the populations
    """
    
    element_class = BasePopulation
    default_size = 2

    params = {'migrate_prob': 0.75}

    alias = {'populations': 'elements',
    'n_populations': 'n_elements',
    'best_population': 'best_element',
    'worst_population': 'worst_element',
    'get_best_population': 'get_best_element',
    'get_best_populations': 'get_best_elements'
    }

    def __str__(self):
        return '\n\n'.join(map(str, self))

    def migrate(self, migrate_prob=None, copy=True):
        migrate_prob = migrate_prob or self.migrate_prob
        for population, other in zip(self[:-1], self[1:]):
            if random() < migrate_prob:
                other.append(population.get_best_individual(copy=copy))
                population.append(other.get_best_individual(copy=copy))

    def transition(self, *args, **kwargs):
        self.migrate()
        for p in self:
            p.transition(*args, **kwargs)

    @side_effect
    def select(self):
        for p in self:
            p.select()

    @side_effect
    def mutate(self):
        for p in self:
            p.mutate()

    def mate(self):
        raise NotImplementedError

    # def max_fitness(self):
    #     return max(map(attrgetter('max_fitness'), self))

    def get_best_individual(self, copy=True):
        bests = map(methodcaller('get_best_individual'), self)
        k = max(b.fitness for b in bests)
        if copy:
            return bests[k].copy()
        else:
            return bests[k]

    @property
    def individuals(self):
        return list(concat(map(attrgetter('individuals'), self)))


class BaseCommunity(BaseMultiPopulation):
    # An alias of `MultiPopulation`

    def __str__(self):
        return ' @\n\n'.join(map(str, self))


class BaseEnvironment(CollectiveMixin, metaclass=MetaContainer):

    """Base Class of environments

    The main method is `evaluate`, computing the fitness of an individual or a population
    """

    element_class = None

    def __init__(self, elements=[]):
        for e in elements:
            e.environment = self

    def evaluate(self, *args, **kwargs):
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return evaluate(self, *args, **kwargs)

    def select(self, pop, n_sel):
        raise NotImplementedError

    def __enter__(self, *args, **kwargs):
        globals()['_fitness'] = lambda o: self._evaluate(o.decode())
        globals()['_environment'] = self
        return self

    def __exit__(self, *args, **kwargs):
        if '_fitness' in globals():
            del globals()['_fitness']
        if '_environment' in globals():
            del globals()['_environment']




#!/usr/bin/env python3

"""GA for optimization of neural networks
"""

import numpy as np

from sklearn.neural_network import MLPRegressor

from .. import MixedIndividual, FloatChromosome, FloatMatrixChromosome
from ..population import StandardPopulation
from ..learn import BaseEstimator


class GAMLPRegressor(BaseEstimator, MLPRegressor):
    """GA for MLP Regression
    """

    hidden_dim = 4
    max_iter = 100
    n_layers = 3

    estimated_params = ('coefs_', 'intercepts_')

    @classmethod
    def create_model(cls, *args, **kwargs):
        # create MLPRegressor object
        model = MLPRegressor(hidden_layer_sizes=(cls.hidden_dim,), max_iter=1, *args, **kwargs)
        model.out_activation_ = 'identity'
        model.n_layers_ = cls.n_layers
        return model

    def __init__(self, *args, **kwargs):
        # create MLPRegressor
        super().__init__(hidden_layer_sizes=(self.hidden_dim,), max_iter=1, *args, **kwargs)
        self.out_activation_ = 'identity'
        self.n_layers_ = 3

    @classmethod
    def config(cls, X, Y, n_individuals=10, *args, **kwargs):
        # configure the population for GA

        input_dim = X.shape[1]
        output_dim = Y.shape[1]

        class MyIndividual(MixedIndividual):

            element_class = FloatMatrixChromosome, FloatChromosome, FloatMatrixChromosome, FloatChromosome

            def _fitness(self):
                model = self.decode()
                return model.score(X, Y)

            def decode(self):
                model = cls.create_model(*args, **kwargs)
                model.coefs_ = tuple(map(np.asarray, self[::2]))
                model.intercepts_ = tuple(map(np.asarray, self[1::2]))
                model.n_layers_ = 3
                return model

        MyPopulation = StandardPopulation[MyIndividual]

        return MyPopulation.random(n_individuals=n_individuals, size=((input_dim, cls.hidden_dim), cls.hidden_dim, (cls.hidden_dim, output_dim), output_dim))




#!/usr/bin/env python3

import numpy as np

from ..benchmarks import Problem


class Kantorovich(Problem):
    '''
    a: a
    b: b
    '''
    def __init__(self, a=0.5, b=1):
        self.a = a
        self.b = b
        self.x = np.linspace(a, b, 100)

    def __call__(self, p):
        # assert np.sum(p) == 1
        return np.dot(self.x, p) * np.dot(1/self.x, p)


#!/usr/bin/env python3

"""Parallel computing for GAs
"""


import threading
from dask import compute, delayed


class ThreadWithResult(threading.Thread):
    
    def __init__(self, func, *args, **kwargs):
        super().__init__() 
        self._result = None
        self._func = func
        self._args = args
        self._kwargs

    def run(self):
        self._result = self._func(*self._args, **self._kwargs)

    @property
    def result(self):
        return self._result


class MTApply:
    # The multi-threading version of the method `apply`

    def __init__(self, type_=None):
        self.type_ = type_

    def __get__(self, instance, owner):
        def _map(f, type_=None):
            threads = [ThreadWithResult(f, e) for e in instance]
            for p in threads:
                p.start()
            for p in threads:
                p.join()
            type_ = type_ or self.type_
            if type_ is None:
                return (thread.result if hasattr(thread, 'result') else None for thread in threads)
            else:
                return type_([thread.result if hasattr(thread, 'result') else None for thread in threads])
        return _map


class MTMap:
    # The multi-threading version of the inbuilt function `map`

    def __init__(self, type_=None):
        self.type_ = type_

    def __call__(self, f, obj, type_=None):
        threads = [ThreadWithResult(f, e) for e in obj]
        for p in threads:
            p.start()
        for p in threads:
            p.join()
        type_ = type_ or self.type_
        if type_ is None:
            return (thread.result if hasattr(thread, 'result') else None for thread in threads)
        else:
            return type_([thread.result if hasattr(thread, 'result') else None for thread in threads])


class DaskApply:
    # The parallel version of the method `apply` by dask

    def __init__(self, type_=None):
        self.type_ = type_

    def __get__(self, instance, owner):
        def _map(f, type_=None):
            results = compute(*(delayed(f)(e) for e in instance))
            type_ = type_ or self.type_
            if type_ is None:
                return results
            else:
                return type_(results)
        return _map


class DaskMap:
    # The parallel version of the inbuilt function `map` by dask

    def __init__(self, type_=None):
        self.type_ = type_

    def __call__(self, f, obj, type_=None):
        results = compute(*(delayed(f)(e) for e in obj))
        type_ = type_ or self.type_
        if type_ is None:
            return results
        else:
            return type_(results)


def mt_apply(cls):
    cls.apply = MTApply(type_=list)
    return cls


def dask_apply(cls):
    cls.apply = DaskApply(type_=list)
    return cls


if __name__ == '__main__':
    
    @dask_apply
    class C:

        def __init__(self, v=[1,2,3]):
            self.v = v
            self._map = MTMap(type_=tuple)

        @property
        def map(self):
            return self._map

        @map.setter
        def map(self, v):
            self._map = v

        def __iter__(self):
            return iter(self.v)

    c = C([1,2,3,4])
    print(c.map)
    print(c.map(lambda x:x+1, c))
    print(c.apply(lambda x:x+2))
    c.map = DaskMap(type_=tuple)
    print(c.map)
    print(c.map(lambda x:x+1, c))
    c.map = map
    print(c.map)
    print(tuple(c.map(lambda x:x+1, c)))


#!/usr/bin/env python3

from sklearn.base import BaseEstimator as BE

import warnings, os
warnings.filterwarnings("ignore")
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'


class BaseEstimator(BE):

    """Base class for machine learning by GA
    
    Attributes:
        estimated_params (tuple): estimtated/learnable parameters by GA
        pop (Population): the population for GA
    """
    
    pop = None
    estimated_params = ()

    @classmethod
    def config(cls, X, Y=None, *args, **kwargs):
        # configure a population for GA
        raise NotImplementedError

    def fit(self, X, Y=None, pop=None, warm_start=False):
        if warm_start:
            self.pop = pop or self.pop or self.config(X, Y)
        else:
            self.pop = pop or self.config(X, Y)
        self._fit()
        return self

    def _fit(self):
        self.pop.ezolve(n_iter=self.max_iter)
        model_ = self.pop.solution
        for k in self.estimated_params:
            setattr(self, k, getattr(model_, k))


# def ga_opt(model_class):
#     class cls(BaseEstimator, model_class):
#         @classmethod
#         def create_model(cls, *args, **kwargs):
#             return model_class(*args, **kwargs)

#     return cls




#!/usr/bin/env python3

"""Ant Colony Optimization

*Ref*
Blum, Christian. "Ant colony optimization: Introduction and recent trends." Physics of Life reviews 2.4 (2005): 353-373.
"""


from types import MethodType
import numpy as np
from .base import FitnessMixin, PopulationMixin
from .meta import MetaContainer
from scipy.stats import rv_discrete

from random import random


class BaseAnt(FitnessMixin):

    initial_position = 0

    params = {"initial_position": None,
    "path": None,
    "n_steps": 1}

    def init(self):
        self.path = [0]

    def __init__(self, *args, **kwargs):
        if args:
            raise Exception('__init__ has no position arguments!')
        for k, v in kwargs.items():
            setattr(self, k, v)

    def move(self, colony=None, n_steps=1):
        colony = colony or self._colony
        i = self.path[-1]
        self.new_path = [i]
        for _ in range(n_steps):
            next_positions = [j for j in colony.positions if j not in self.path]
            if not next_positions: break
            elif len(next_positions) == 1:
                i = next_positions[0]
            else:
                ps = np.array([colony.move_proba[i, j] for j in colony.positions if j not in self.path])
                ps += 0.0001
                ps /= np.sum(ps)
                rv = rv_discrete(values=(next_positions, ps))
                i = rv.rvs(size=1)[0]
            self.new_path.append(i)
            self.path.append(i)
        if len(self.new_path) >= 2:
            self.release_pheromone(colony)

    def release_pheromone(self, colony):
        self.pheromone = colony.sedimentation / np.sum(colony.distances[i,j] for i, j in zip(self.new_path[:-1], self.new_path[1:]))

    def get_length(self, distances=None):
        distances = distances or self._colony.distances
        return np.sum(distances[i,j] for i, j in zip(self.path[:-1], self.path[1:]))

    def _fitness(self):
        n = len(self._colony.positions) - len(self.path)
        if n == 0:
            return 1 / self.get_length()
        else:
            rest_positions = [self.path[-1]] + [j for j in self._colony.positions if j not in self.path]
            M = np.mean([self._colony.distances[i,j] for i in rest_positions for j in rest_positions if i<j])
            return 1 / (self.get_length() + n*M)

    def reset(self):
        self.path = [0]
        

class BaseAntColony(PopulationMixin, metaclass=MetaContainer):

    element_class = BaseAnt
    params = {'sedimentation':100, 'volatilization':0.75, 'alpha':1, 'beta':5, 'n_steps':1}

    alias = {"ants": "elements", "worst_ant": "worst_element", "get_worst_individuals": "get_worst_elements"}

    @classmethod
    def from_positions(cls, n_ants=10, positions=None):
        from scipy.spatial.distance import pdist, squareform
        obj = cls([cls.element_class() for _ in range(n_ants)])
        obj.positions = positions
        obj.pheromone = np.zeros((len(positions),)*2)
        obj.distances = squareform(pdist(points))
        obj.observe()
        return obj

    @classmethod
    def from_distances(cls, n_ants=10, distances=None):
        obj = cls([cls.element_class() for _ in range(n_ants)])
        obj.positions = np.arange(distances.shape[0])
        obj.distances = distances
        obj.pheromone = np.zeros(distances.shape)
        obj.observe()
        return obj

    def observe(self):
        for ant in self:
            ant._colony = self

    def transition(self, *args, **kwargs):
        self.move(n_steps=self.n_steps)
        self.update_pheromone()
        for ant in self.get_worst_individuals(0.3):
            ant.reset()

    @property
    def move_proba(self):
        return self.pheromone**self.alpha / np.maximum(self.distances**self.beta, 0.1)

    def init(self):
        for ant in self:
            ant.init()

    def move(self, *args, **kwargs):
        for ant in self:
            ant.move(self, *args, **kwargs)

    def update_pheromone(self):
        delta = np.zeros_like(self.pheromone)
        for ant in self:
            for (i, j) in zip(ant.new_path[:-1], ant.new_path[1:]):
                delta[i, j] += ant.pheromone
        self.pheromone = (1- self.volatilization) * self.pheromone + delta




#!/usr/bin/env python3

"""Particle Swarm Optimization

Developed by J. Kennedy and R. Eberhart[Kennedy and Eberhart 2001]

Each individual is represented by the position and the velocity.
"""

from random import random
from operator import attrgetter

import numpy as np

from .base import BaseIndividual
from .chromosome import FloatChromosome
from .mixin import PopulationMixin
from .meta import MetaContainer
from .deco import basic_memory


@basic_memory
class BaseParticle(BaseIndividual):
    """A particle in PSO

    An individual represented by 2 chromosomes: position and velocity
    
    Variables:
        default_size {number} -- 2 by default
        memory {Particle} -- the best state of the particle moving in the solution space.

    Caution:
        Have to implement the properties in subclasses: position, velocity
    """

    element_class = FloatChromosome
    default_size = 2

    @property
    def position(self):
        raise NotImplementedError

    @position.setter
    def position(self, x):
        raise NotImplementedError

    @property
    def velocity(self):
        raise NotImplementedError

    @velocity.setter
    def velocity(self, v):
        raise NotImplementedError

    @property
    def best_position(self):
        # alias for the position of memory
        if self.memory.get('solution', None) is None:
            return self.position
        return self.memory['solution']

    def update_vilocity(self, fame=None, *args, **kwargs):
        raise NotImplementedError

    def move(self, velocity=None):
        if velocity is None:
            self.position += self.velocity
        else:
            self.position += velocity

    def decode(self):
        return self.best_position


class Particle(BaseParticle):
    """
    Standard Particle

    chromosomes = (position, velocity)
    """

    element_class = FloatChromosome
    default_size = 2

    params = {'learning_factor': 2,
        'acceleration_coefficient': 3,
        'inertia': 0.5
        }

    @property
    def position(self):
        return self.chromosomes[0]

    @position.setter
    def position(self, x):
        self.chromosomes[0] = x
        self.after_setter()

    @property
    def velocity(self):
        return self.chromosomes[1]

    @velocity.setter
    def velocity(self, v):
        self.chromosomes[1] = v

    @property
    def direction(self):
        return self.best_position - self.position

    def update_vilocity(self, scale, inertia, learning_factor):      
        self.velocity = (inertia * self.velocity
                        + learning_factor * scale * self.direction)
    
    def update_vilocity_by_fame(self, fame, scale, scale_fame, inertia, learning_factor, acceleration_coefficient):
        self.velocity = (inertia * self.velocity
                        + learning_factor * scale * self.direction
                        + acceleration_coefficient * scale_fame * (fame.best_position-self.position))


class ParticleSwarm(PopulationMixin, metaclass=MetaContainer):
    """Standard PSO
    
    Extends:
        PopulationMixin
    """
    
    element_class = Particle
    default_size = 20

    params = {'learning_factor': 2, 'acceleration_coefficient': 3,
    'inertia':0.75, 'hof_size':0.2, 'max_velocity':None}

    alias = {
    'particles': 'elements',
    'n_particles': 'n_elements',
    'best_particle': 'best_element',
    'get_best_particle': 'get_best_element',
    'get_best_particles': 'get_best_elements'
    }

    def init(self):
        super().init()
        self.hall_of_fame = self.get_best_particles(self.hof_size, copy=True)
    
    def update_hall_of_fame(self):
        hof_size = len(self.hall_of_fame)
        for ind in self:
            for k in range(hof_size):
                if self.hall_of_fame[-k-1].fitness < ind.fitness:
                    self.hall_of_fame.insert(hof_size-k, ind.copy())
                    self.hall_of_fame.pop(0)
                    break

    @property
    def best_fitness(self):
        if self.hall_of_fame:
            return max(map(attrgetter('fitness'), self.hall_of_fame))
        else:
            return super().best_fitness

    def transition(self, *args, **kwargs):
        """
        Transitation of the states of particles
        """
        self.move()
        self.backup()
        self.update_hall_of_fame()

    def backup(self):
        # overwrite the memory of the particle if its current state is better its memory
        for particle in self:
            particle.backup()

    def move(self):
        """Move the particles

        Define the moving rule of particles, according to the hall of fame and the best record
        """

        scale = random()
        eta = random()
        scale_fame = random()
        for particle in self:
            for fame in self.hall_of_fame:
                if particle.fitness < fame.fitness:
                    particle.update_vilocity_by_fame(fame, scale, scale_fame, 
                        self.inertia, self.learning_factor, self.acceleration_coefficient)
                    particle.position = particle.position + particle.velocity
                    break
        for particle in self.hall_of_fame:
            particle.update_vilocity(scale, self.inertia, self.learning_factor)
            particle.position = particle.position + particle.velocity


class DiscreteParticleSwarm(ParticleSwarm):
    # PSO with discrete particles

    def move(self):
        """
        Move the discrete particles
        """

        v1, v2, v3 = (self.inertia,
            self.learning_factor,
            self.acceleration_coefficient)
        for particle in self:
            if particle in self.hall_of_fame:
                particle.velocity = _velocity1(particle, v1, v2)
            else:
                for fame in self.hall_of_fame:
                    if particle.fitness < fame.fitness:
                        break
                particle.velocity = _velocity2(particle, fame, v1, v2, v3)
            particle.move()



#!/usr/bin/env python3


"""
Simulated Annealing Algorithm

*Ref*
S. Kirkpatrick, C. D. Gelatt, Jr., M. P. Vecchi. Optimization by Simulated Annealing. 1983: 220(4598): 671-679
"""


from .. import PhantomIndividual
from .. import metropolis_rule


class SimulatedAnnealing(PhantomIndividual):
    """Class for Simulated Annealing
    
    Attributes:
        params (dict): parameters in SA
    """

    params = {'ext_c': 0.99,  # external coef
        'int_c': 0.99,        # internal coef
        'n_epochs': 200,
        'initT': 100,         # initial temperature
        'termT': 0.0001       # terminal temperature
        }

    def transition(self, *args, **kwargs):
        T = self.initT
        for epoch in range(self.n_epochs):
            self.move(T)
            T *= self.int_c
            if T < self.termT:
                break
        # set the phantom to be the true solution (if it is better then the previous record)
        self.backup()
        self.initT = T * self.ext_c

    def move(self, T):
        """Move phantom
        
        Arguments:
            T {number} -- temperature
        """

        cpy = self.phantom.get_neighbour()

        # Metropolis rule
        flag = metropolis_rule(D=cpy.fitness - self.phantom.fitness, T=T)
        if flag:
            self.phantom.chromosomes = cpy.chromosomes




#!/usr/bin/env python3

"""
Individual classes

An individual represents a solution of a optimization problem.
It is defined as a container of chromosomes.
"""

from .base import BaseIndividual
from .chromosome import BinaryChromosome, BaseChromosome, FloatChromosome
from .meta import MetaTuple, MetaList, MetaSingle
from .utils import randint


class MultiIndividual(BaseIndividual, metaclass=MetaList):
    pass


PolyIndividual = MultiIndividual


class MonoIndividual(BaseIndividual, metaclass=MetaSingle):
    """Base class of individual with only one chromosome;
    It is equavalent to a chromosome.

    You should implement the genetic operations: cross, mutate.
    """

    @classmethod
    def random(cls, *args, **kwargs):
        return cls([cls.element_class.random(*args, **kwargs)])

    @property
    def chromosome(self):
        return self.chromosomes[0]

    @chromosome.setter
    def chromosome(self, v):
        self.chromosomes[0] = v

    def decode(self):
        return self.chromosome.decode()

    @classmethod
    def set_size(cls, sz):
        raise DeprecationWarning("Never set the size of this class!")


class MixedIndividual(BaseIndividual, metaclass=MetaTuple):
    """base class of individual

    You should implement the enetic operations: cross, mutate
    """

    element_class = BaseChromosome, BaseChromosome

    @property
    def default_size(self):
        return len(self.element_class)

    @classmethod
    def random(cls, size=None, n_chromosomes=None, *args, **kwargs):
        if size is None:
            return cls([C.random(*args, **kwargs) for C in cls.element_class])
        elif isinstance(size, int):
            size = (size,) * n_chromosomes
        else: #if isinstance(size, tuple):
            if len(size) != len(cls.element_class):
                raise ValueError('the length of `size` is not equal to the number of elements (`n_chromosomes`)')
        return cls([C.random(size=l, *args, **kwargs) for C, l in zip(cls.element_class, size)])


class AgeIndividual(BaseIndividual):

    params = {
    "age": 0,
    "life_span": 100
    }


class GenderIndividual(MixedIndividual):

    @property
    def gender(self):
        raise NotImplementedError


from .deco import fitness_cache


@fitness_cache
class PhantomIndividual(BaseIndividual):
    # Anthor way to implement memory system for individual

    phantom = None

    def init(self):
        self.phantom = self.copy()

    def backup(self):
        if self.fitness < self.phantom.fitness:
            self.chromosomes = self.phantom.chromosomes
            self.set_cache(fitness=self.phantom.fitness)


# Following are functions to create individuals

def binaryIndividual(size=8):
    """simple binary individual
    encoded as a sequence such as 01001101

    Equiv. to `makeIndividual(size=size)`
    """

    return MonoIndividual[BinaryChromosome.set(default_size=size)]


def makeIndividual(element_class=BinaryChromosome, n_chromosomes=None, size=8, cls=None):
    """helper to make an individual
    
    Example:
        # make an indiviudal with 2 binary chromosomes, each chromosome has 8 genes,
        makeIndividual(element_class=BinaryChromosome, n_chromosomes=2, size=8)
        # make mixed indiviudal with 2 type of chromosomes, one has 8 genes and the other has 2,
        makeIndividual(element_class=(BinaryChromosome, FloatChromosome), size=(8,2))
    
    Args:
        element_class (BaseChromosome, tuple, optional): class of chromosomes
        n_chromosomes (int, optional): the number of chromosomes
        size (int, tuple, optional): the sizes of chromosomes
        cls: the class of the return individual
    
    Returns:
        cls or other individual classes
    """

    if n_chromosomes is None:
        if isinstance(size, tuple):
            n_chromosomes = len(size)
        else:
            n_chromosomes = 1

    if n_chromosomes == 1:
        assert isinstance(size, int)
        cls = cls or MonoIndividual
        return cls[element_class // size]
    else:
        if isinstance(size, tuple):
            if len(size) == n_chromosomes:
                if isinstance(element_class, tuple):
                    cls = cls or MixedIndividual
                    return cls[tuple(e_c // s for e_c, s in zip(element_class, size))]
                else:
                    cls = cls or MixedIndividual
                    return cls[tuple(element_class // s for s in size)]
            else: 
                raise ValueError('the length of `size` must be `n_chromosomes`.')
        elif isinstance(size, int):
            cls = cls or PolyIndividual
            return cls[element_class // size].set(default_size=n_chromosomes)
        else:
            raise ValueError('the length of `size` must be a number or a tuple of numbers.')


def makeBinaryIndividual(size=8, cls=None):
    cls = cls or MixedIndividual
    if isinstance(size, tuple):
        return cls[tuple(BinaryChromosome // s for s in size)]
    else:
        return cls[BinaryChromosome // size]




#!/usr/bin/env python3


class BaseProblem:
    """Base Class of Problems

    The class is just a function (callable object).
    Please implement `__call__` method in each subclass
    make it behaving as function

    In fact, it is never used, just a template.
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError


#!/usr/bin/env python3

"""Sparrow Search Algorithm

Algorithm 1 The framework of the SSA. 
Input: 
G: the maximum iterations 
PD: the number of producers 
SD: the number of sparrows who perceive the danger R2 : the alarm value 
n: the number of sparrows 
Initialize a population of n sparrows and define its relevant parameters. 
Output: Xbest, fg. 
while (t < G) 
    Rank the fitness values and find the current best individual and the current worst individual. 
    R2 = rand(1)
    for i = 1 : PD
        update the sparrow’s location;
    for i = (PD + 1) : n
        update the sparrow’s location;
    for l = 1 : SD
        update the sparrow’s location;
    Get the current new location;
    If the new location is better than before, update it;
    t = t + 1
return Xbest, fg.

*References*
Jiankai Xuea, and Bo Shena, A novel swarm intelligence optimization approach: sparrow search algorithm.
"""

from random import gauss, random, randint
import numpy as np
from scipy.spatial.distance import pdist, squareform

from .base import PopulationMixin
from .chromosome import FloatChromosome
from .individual import PolyIndividual


@basic_memory
class BaseSparrow(FloatChromosome):

    def move(self, ST, i):
        r = random()
        Q = gauss()
        alpha = random()+0.01
        if r<ST:
            self *= np.exp(-i/(T*alpha))
        else:
            self += Q


class Producer(BaseSparrow):
    pass


class Scrounger(BaseSparrow):

    def move(self, worst, producer, i):
        r = random()
        if r < 0.5:
            Q = gauss()
            self[:] = Q * np.exp((self - worst) / i**2)
        else:
            d = len(self)
            self[:] = producer + np.abs(self - producer) * (np.random.randint(2, size=d)*2 - 1) /d**2


class StandardSparrowSearch:
    """Starndard Sparrow Search Algorithm
    """

    element_class = BaseSparrow

    params = {
        "PD": 0.2,
        "SD": 0.2,
        "ST": 0.7,
        'rho': 0.001
    }

    def init(self):
        if 0< self.PD <1:
            self.PD = int(self.PD * len(self))
        if 0< self.SD <1:
            self.SD = int(self.SD * len(self))

    def transition(self, *args, **kwargs):

        self.sort()

        producers = self[:self.PD]
        
        for i, sparrow in enumerate(producers):
            sparrow.move(self.ST, i)

        k = np.argmax(list(map(lambda x: x.fitness, producers)))
        producer = producers[k]

        for i, sparrow in enumerate(self[self.PD:]):
            Scrounger.move(sparrow, self[0], producer, i+self.PD)

        best, worst = self[-1], self[0]
        bf, wf = self.max_fitness, self.min_fitness

        for sparrow in self.random_select(n_sel=self.SD):
            if sparrow.fitness < bf:
                beta = gauss()
                sparrow[:] = best + beta * np.abs(best - sparrow)
            elif sparrow.fitness == bf:
                K = randint(0, 1)*2 - 1
                sparrow += K * np.abs(sparrow - worst) / (sparrow.fitness - wf + self.rho)

       self.update()

